#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/md5.h>
#include <openssl/rsa.h>
#include <openssl/objects.h>

#define N		4
#define KEY_S	16
#define ROUNDS	100
#define BUF_MAX	1024*1024

#define toLL(tv) ((tv.tv_sec*(long long)1000000)+tv.tv_usec)

int main() {
	
	int i, j, x;
	MD5_CTX md5_ctx;
	RSA *rsa;
	u_char *sig;
	u_int siglen;
	u_char key[KEY_S];
	u_char *buf;
	u_int buf_s;
	u_char hash[MD5_DIGEST_LENGTH];
	u_char haux[MD5_DIGEST_LENGTH];
	struct timeval t0, t1;
	
	for(buf_s = 8 ; buf_s <= BUF_MAX ; buf_s *= 2) {
		buf = malloc(buf_s);
////////////////////////////////////////////////////////////////////////////////	
	
	gettimeofday(&t0, NULL);
	for(i = 0 ; i < ROUNDS ; i++) {
		for(j = 0 ; j < N ; j++) {
			MD5_Init(&md5_ctx);
			MD5_Update(&md5_ctx, buf, buf_s);
			MD5_Update(&md5_ctx, key, KEY_S);
			MD5_Final(hash, &md5_ctx);
		}
	}
	gettimeofday(&t1, NULL);
	printf("H(M | K) \t%d bytes, %d rounds, %d us\n", buf_s, ROUNDS, toLL(t1) -
toLL(t0));

////////////////////////////////////////////////////////////////////////////////

	gettimeofday(&t0, NULL);
	for(i = 0 ; i < ROUNDS ; i++) {
		MD5(buf, buf_s, haux);
		for(j = 0 ; j < N ; j++) {
			MD5(key, KEY_S, haux);
		}
	}
	gettimeofday(&t1, NULL);
	printf("H(H(M) | K) \t%d bytes, %d rounds, %d us\n", buf_s, ROUNDS, toLL(t1)
- toLL(t0));

////////////////////////////////////////////////////////////////////////////////
	
	//rsa = RSA_new();
	rsa = RSA_generate_key(1024, 17, NULL, NULL);
	if(!rsa) {
		perror("rsa");
		exit(0);
	}

	MD5_Init(&md5_ctx);
	MD5_Update(&md5_ctx, buf, buf_s); 
	MD5_Update(&md5_ctx, key, KEY_S); MD5_Final(hash, &md5_ctx);
	
	sig = malloc(RSA_size(rsa));

	gettimeofday(&t0, NULL);
	for(i = 0 ; i < ROUNDS ; i++) {
		RSA_sign(NID_md5, hash, MD5_DIGEST_LENGTH, sig, &siglen, rsa);
	}
	gettimeofday(&t1, NULL);	
	
	RSA_free(rsa);
	free(sig);
	
	printf("S(H(M)) \t%d bytes, %d rounds, %d us\n", buf_s, ROUNDS, toLL(t1) -
toLL(t0));	

////////////////////////////////////////////////////////////////////////////////

	printf("\n");
	free(buf);
	}
}
