/**
 * Benchmark client for RITAS.
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "benchmark.h"


#define toLL(tv) ((tv.tv_sec*(long long)1000000)+tv.tv_usec)

void usage(char *argv) {
	printf("Usage: %s < rb | bc | mvc | vc | ab <sender> <rate> <msg_size> > <req> <procfile>\n", argv);
	printf(" rb - Reliable Broadcast\n");
	printf(" bc - Binary Consensus\n");
	printf("mvc - Multivalued Consensus\n");
	printf(" vc - Vector Consensus\n");
	printf(" ab - Atomic Broadcast\n");
	printf("<sender> - ID of sender; -1 makes all processes senders\n");
	printf("<rate> - Milliseconds between requests for Atomic Broadcast\n");
	printf("<req> - Number of Requests\n");
	printf("<procfile> - File processes' addresses\n");
}

int main(int argc, char **argv) {

	u_char protocol;
	int rate, requests;
	int sock, tcpsock;
	FILE *fp;
	int i, j, n, procn;
	char buf[BUF_S];
	char *s;
	struct sockaddr_in *addr, servaddr, cliaddr;
	socklen_t cliaddr_s;
	void *msg;
	u_char msg_size;
	socklen_t addr_s;
	struct linger lg;
	
	if(argc < 4) {
		usage(argv[0]);
		exit(1);
	}
	
	/*
	 * Parse command line arguments.
	 */
	requests = atoi(argv[2]);
	n = 3;
	if(!memcmp(argv[1], "rb", 2)) {
		protocol = RB;

		msg = malloc(sizeof(struct rb_bench));
		((struct bc_bench *)msg)->protocol = RB;
		((struct rb_bench *)msg)->sender = 0;
		((struct rb_bench *)msg)->value_s = VALUE_S;
		msg_size = sizeof(struct rb_bench);
	}
	else if(!memcmp(argv[1], "bc", 2)) {
		protocol = BC;

		msg = malloc(sizeof(struct bc_bench));
		((struct bc_bench *)msg)->protocol = BC;
		((struct bc_bench *)msg)->requests = requests;
		msg_size = sizeof(struct bc_bench);
	}
	else if(!memcmp(argv[1], "mvc", 3)) {
		protocol = MVC;

		msg = malloc(sizeof(struct mvc_bench));
		((struct mvc_bench *)msg)->protocol = MVC;
		((struct mvc_bench *)msg)->value_s = VALUE_S;
		msg_size = sizeof(struct mvc_bench);		
	}
	else if(!memcmp(argv[1], "vc", 2)) {
		protocol = VC;

		msg = malloc(sizeof(struct vc_bench));
		((struct vc_bench *)msg)->protocol = VC;
		((struct vc_bench *)msg)->value_s = VALUE_S;
		msg_size = sizeof(struct vc_bench);
	}
	else if(!memcmp(argv[1], "ab", 2)) {
		if(argc < 7) {
			usage(argv[0]);
			exit(1);
		}		
		protocol = AB;
		rate = atoi(argv[3]);
		requests = atoi(argv[5]);
		n = 6;

		msg = malloc(sizeof(struct ab_bench));		
		((struct ab_bench *)msg)->protocol = AB;
		((struct ab_bench *)msg)->sender = atoi(argv[2]);
		((struct ab_bench *)msg)->requests = requests;		
		((struct ab_bench *)msg)->rate = rate;
		((struct ab_bench *)msg)->value_s = atoi(argv[4]);
		msg_size = sizeof(struct ab_bench);
	}
	else {
		usage(argv[0]);
		exit(1);
	}
	
	/*
	 * Force unbuffered output.
	 */
	setlinebuf(stdout);

	/*
	 * Open and parse processes' file
	 */
	fp = fopen(argv[n], "r");
	if(!fp) {
		perror("fopen");
		exit(1);
	}

	s = fgets(buf, BUF_S, fp);
	if(!s) {
		perror("fgets");
		exit(1);
	}

	procn = atoi(buf);
	if(procn < 4 || procn > 255) {
		printf("invalid number of processes\n");
	}
	
	addr = malloc(procn * sizeof(*addr));
	
	for(i = 0 ; i < procn ; i++) {
		s = fgets(buf, BUF_S, fp);
		if(!s) {
			perror("fgets");
			exit(1);
		}
		buf[strlen(buf)-1] = '\0';
		
		inet_aton(buf, &addr[i].sin_addr);
		
		s = fgets(buf, BUF_S, fp);
		if(!s) {
			perror("fgets");
			exit(1);
		}
		
		addr[i].sin_port = htons(BASE_PORT+i);
		addr[i].sin_family = AF_INET;
	}
	fclose(fp);

	/*
	 * Create socket.
	 */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if(sock < 0) {
		perror("socket");
		exit(1);
	}
/*
	tcpsock = socket(AF_INET, SOCK_STREAM, 0);
	if(tcpsock < 0) {
		perror("socket");
		exit(1);
	}	
	lg.l_onoff = 1;
	lg.l_linger = 0;
	setsockopt(tcpsock, SOL_SOCKET, SO_LINGER, (char *)&lg, sizeof(lg));
	
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons(BASE_PORT);
	n = bind(tcpsock, (struct sockaddr *)&servaddr, sizeof(servaddr));
	if(n < 0) {
		perror("bind");
		exit(1);
	}
	n = listen(tcpsock, 10);
	if(n < 0) {
		perror("listen");
		exit(1);
	}
*/
	/*
	 * Send the signal to all processes.
	 */
	//for(j = 0 ; j < requests ; j++) {
		for(i = 0 ; i < procn ; i++) {
			((struct bench *)msg)->id = 1;//j+1;
			n = sendto(sock, msg, msg_size, 0, (struct sockaddr *)&addr[i], sizeof(addr[i]));
			if(n < 0) {
				perror("sendto");
				exit(1);
			}
		}
		if(protocol == AB) {
			//break;
		}
		else {
			sleep(1);
			//tv.tv_sec = 0;
			//tv.tv_usec = 3000;
			//select(0, NULL, NULL, NULL, &tv);
		}
	//}
	
	/*
	 * Wait for all processes to complete.
	 */
/*
	for(i = 0 ; i < procn ; i++) {
		n = accept(tcpsock, (struct sockaddr *)&cliaddr, &cliaddr_s);
		close(n);
	}
*/	 
	for(i = 0 ; i < procn ; i++) {
		addr_s = sizeof(addr[i]);
		n = recvfrom(sock, buf, BUF_S, 0, (struct sockaddr *)&addr[i], &addr_s);
		if(n < 0) {
			perror("recvfrom");
			exit(1);
		}
	}

	/*
	 * Cleanup.
	 */
	free(addr);
	free(msg);
	close(sock);

	exit(0);
}
