
int	libdit_reliable_broadcast(libdit_t *ctx, char role, char *buf, u_int buf_s){
	
	switch(ctx->rb_type) {
		default:
		case LIBDIT_RB_BRACHA:
			return libdit_rb_bracha(ctx, role, buf, buf_s);
			break;
	}
}

int libdit_rb_bracha(libdit_t *ctx, char role, char *buf, u_int buf_s) {
	
	int i, j;
	struct libdit_msg msg;
	libdit_vlist_t *vlist[3];
	libdit_proc *proc;

	/*
	 * Initialize the value list for each message type: INIT, READY and ECHO.
	 * These lists store the number of times each value is received.
	 */
	for(i = 0 ; i < 3 ; i++) {
		vlist[i] = libdit_vlist_init();
		if(!vlist[i]) {
			/*
			 * Tricky. We have to destroy the lists initialized in
			 * previous iterations before returning, heh.
			 */
			for(j = 0 ; j < i ; j++)
				libdit_vlist_destroy(vlist[j]);
				
			return -1;
		}
	}
	
	
	/*
	 * Step 0 (only by the transmitter).
	 * Broadcast the value to all the processes (including itself).
	 */
	if(role == LIBDIT_RB_SENDER) {
		libdit_broadcast(ctx, buf, buf_s);
	}
	

	/*
	 * Step 1.
	 * Wait till receive for some value,
	 * 	one (LIBDIT_RB_INITIAL, value) message
	 * 	or (n + t)/2 (LIBDIT_RB_ECHO, value) messages,
	 * 	or (t + 1) (LIBDIT_RB_READY, value) messages.
	 * Then broadcast (LIBDIT_RB_ECHO, v) to all the processes.
	 */
	while( libdit_value_count(vlist_init, msg->value, msg->value_size) < 1
		|| libdit_value_count(vlist_echo, msg->value, msg->value_size) <= (n + t)/2
		|| libdit_value_count(vlist_ready, msg->value, msg->value_size) < t + 1
		)
	{
		
		n = libdit_recv(ctx, &buf, &size);
		if(n < 0) {
			libdit_error(ctx, ERROR_MSG);
			return -1;
		}
		
		libdit_rb_extract(&msg, buf, buf_s);
		
		/*
		 * See if it's the first message received from this particular process
		 * with this msg.type value.
		 * This may seem pedantic, but any other way is potentialy dangerous.
		 */
		i = -1;
		switch(msg.type) {
			case LIBDIT_RB_INIT:
				if(!proc->rb_recv_init) {
					proc->rb_recv_init = 1;
					i = 0;
				}
				break;
			case LIBDIT_RB_ECHO:
				if(!proc->rb_recv_echo) {
					proc->rb_recv_echo = 1;
					i = 1;
				}
				break;
			case LIBDIT_RB_READY:
				if(!proc->rb_recv_ready) {
					proc->rb_recv_ready = 1;
					i = 2;
				}
				break;
		}
		/*
		 * It's cool. Now let's amp up the count for this value.
		 */
		if(i != -1)
			libdit_value_store(vlist[i], &msg);

	}
	libdit_rb_build(buf, buf_s, LIBDIT_RB_ECHO, msg.value, msg.value_size);
	libdit_broadcast(ctx, buf, buf_s);
	

	/*
	 * Step 2.
	 * Wait till receive for some value,
	 * 	(n + t)/2 (LIBDIT_RB_ECHO, value) messages,
	 * 	or (t + 1) (LIBDIT_RB_READY, value) messages.
	 * Then broadcast (LIBDIT_RB_READY, v) to all the processes.
	 */
	while( libdit_msg_count_value(&msg_list, LIBDIT_RB_ECHO, msg.value, 
			msg.value_size) <= (n + t)/2
		|| libdit_msg_count_value(&msg_list, LIBDIT_RB_READY, msg.value, 
			msg.value_size) < t + 1
		)
	{
		
		i = libdit_recv(ctx, &msg);
		if(i < 0) {
			libdit_error(ctx, ERROR_MSG);
			return -1;
		}
		
		i = libdit_msg_count_from(&msg_list, msg.type, msg.proc_id);
		if(!i) {
			libdit_msg_add(&msg_list, &msg);
		}
		
	}
	libdit_broadcast(ctx, msg.value, msg.value_size);

	/*
	 * Step 3.
	 * Wait till receive for some value,
	 * 	(2t + 1) (LIBDIT_RB_READY, value) messages.
	 * We accept the value.
	 */
	while( libdit_msg_count_value(&msg_list, LIBDIT_RB_READY, msg.value, 
			msg.value_size) < 2t + 1
		)
	{
		
		i = libdit_recv(ctx, &msg);
		if(i < 0) {
			libdit_error(ctx, ERROR_MSG);
			return -1;
		}
		
		i = libdit_msg_count_from(&msg_list, msg.type, msg.proc_id);
		if(!i) {
			libdit_msg_add(&msg_list, &msg);
		}
		
	}

	memcpy(buf, msg.value, msg.value_size);	

	libdit_vlist_destroy(vlist_init);
	libdit_vlist_destroy(vlist_echo);
	libdit_vlist_destroy(vlist_ready);		

	return msg.value_size;
	