/*
 * The context control structure over ritas's operation.
 * Contains all the necessary data to maintain state on an intrusion-tolerant 
 * communication. This structure is completely opaque to the programmer via the
 * ritas_t datatype, its internal contents should never be needed to be 
 * directly accessed.
 */
struct ritas_context {
	
	char error[RITAS_ERRBUF_SIZE];
#define ritas_ERRBUF_SIZE 64
	int socket;
	struct ritas_group_list *group;
	u_short group_size;

	struct ritas_queue_head **queue_table;
	u_int queue_table_size;
	
};	
typedef struct ritas_context ritas_t;

/*
 * Group management structure.
 * Maintains information such as the network
 * addresses for each process belonging to the group.
 */
struct ritas_group_list {
	struct sockaddr_in addr;
	int rb_echo_recv;
	int rb_init_recv;
	int rb_ready_recv;
	struct ritas_group_list *next;
};
typedef struct ritas_group_list ritas_proc_t;


// RITAS MBUF
struct ritas_mbuf {
	char *rc_h;
	char *rb_h;
	char *payload;
	u_short payload_s;
	struct ritas_mbuf *next;
	char *data;
};
typedef struct ritas_mbuf ritas_mbuf_t;




/**
 * General message structure.
 * Every message used in ritas shares this structure.
 * It contains a process identification, a message type tag,
 * a value and its respective size.
 */
struct ritas_msg {
	u_short proc_id;					/* process identifier */
	u_char msg_type;					/* message type tag */
#define RITAS_RB_INIT	0x00
#define RITAS_RB_ECHO	0x01
#define RITAS_RB_READY	0x02
	ritas_round_t round;				/* round identifier */
	u_int value_size;					/* value size */
	u_char value[RITAS_VALUE_BUF];		/* value carried by the message */
#define RITAS_VALUE_BUF 20				/* 160-bit hashes seem fine! */ 
};
typedef struct ritas_msg ritas_msg_t;

/*
 * Round descriptor.
 */
struct ritas_round {
	u_int ip;
	u_short port;
	u_short bsn;
};
typedef struct ritas_round ritas_round_t;

/*
 * 
 */
struct ritas_queue_head {
	ritas_round_t round;
	ritas_queue_mbuf_t *first;
	ritas_queue_mbuf_t *last;
	struct ritas_queue_head *next;
};

struct ritas_queue_mbuf {
	ritas_msg_t msg;
	ritas_queue_muf_t *next;
};
typedef struct ritas_queue_muf ritas_queue_mbuf_t;
	

/*
 * value list crap.
 */
struct ritas_value_dummy {
	struct ritas_value_count *next;
};
typedef struct ritas_value_dummy ritas_vlist_t;


struct ritas_value_node {
	char *value;
	u_int value_size;
	u_short count;
	struct ritas_value_count *next;
};
typedef struct ritas_value_node ritas_vnode_t;

