#include <stdlib.h>
#include <string.h>
#include "ritas_ht.h"

ritas_ht_t *ritas_ht_init(u_char tsize) {
	
	ritas_ht_t *ht;
	
	ht = malloc(sizeof(ritas_ht_t));
	if(!ht)
		return NULL;
	
	ht->table = malloc(ht->size*sizeof(ritas_ht_entry_t *));
	if(!ht->table) {
		free(ht);
		return NULL;
	}
		
	memset(ht->table, '\0', tsize*sizeof(ritas_ht_entry_t *));
	
	return ht;
}

void ritas_ht_destroy(ritas_ht_t *ht, void (*destroy_func)(void *)) {
	
	int i;
	ritas_ht_entry_t *httmp;
	
	if(!ht)
		return;
		
	for(i = 0 ; i < ht->size ; i++) {
		while(ht->table[i]) {
			httmp = ht->table[i];
			ht->table[i] = ht->table[i]->next;
			
			free(httmp->key);
			
			if(destroy_func)
				(*destroy_func)(ht->table[i]->value);
			else
				free(ht->table[i]->value);
		}
	}
	free(ht->table);
	free(ht);
}


int ritas_ht_insert(ritas_ht_t *ht, u_char *key, u_char ksize, void *object) {
	
	u_char *index_p;
	ritas_ht_entry_t *e = NULL;
	
	index_p = malloc(sizeof(*p));
	if(!index_p)
		goto bad;

		
	*index_p = ritas_ht_hash(key, ksize);
	e = malloc(sizeof(*e));
	if(!e)
		goto bad;
		
	e->object = object;
	e->key = malloc(ksize);
	if(!e->key)
		goto bad;
	memcpy(e->key, key, ksize);
	
	ritas_list_enqueue(ht->table[*index_p], e);
	if(!ht->table[index])
		ritas_list_enqueue(ht->entry_l, index_p);
	else
		free(index_p);

	return 1;
	
bad:
	free(index_p);
	free(e);

	return -1;
}
		
	
	


u_char *ritas_ht_add(ritas_ht_t *ht, u_char *key, u_char ksize, void *value) {
	
	ritas_ht_entry_t *httmp;
	u_char index;
	int i;
	
	for(index = 0, i = 0 ; i < ksize ; i++)
		index += key[i];
	index = index % ht->size;
	
	httmp = malloc(sizeof(ritas_ht_entry_t));
	if(!httmp)
		return 0;
	
	httmp->value = value;
	httmp->key = malloc(ksize);
	httmp->ksize = ksize;
	memcpy(httmp->key, key, ksize);

	if(!ht->table[index])
		httmp->next = NULL;
	else
		httmp->next = ht->table[index];

	ht->table[index] = httmp;

	return httmp->key;	
}


void *ritas_ht_remove(ritas_ht_t *ht, u_char *key, u_char ksize) {

	u_char index;
	ritas_ht_entry_t *httmp, *htprev;
	void *ret;
	int i;
	
	for(index = 0, i = 0 ; i < ksize ; i++)
		index += key[i];
	index = index % ht->size;
	
	htprev = httmp = NULL;
	
		
	for(httmp = ht->table[index] ; httmp ; htprev = httmp, httmp = ht->table[index]->next) {
		if(ksize == httmp->ksize) {
			if(!memcmp(httmp->key, key, ksize)) {
				if(!htprev)
					ht->table[index] = htprev;
				else
					htprev->next = httmp->next;		
				ret = httmp->value;
				free(httmp);
				return ret;
			}
		}
	}
					
	return NULL;	
}


void *ritas_ht_lookup(ritas_ht_t *ht, u_char *key, u_char ksize) {

	u_char index;
	ritas_ht_entry_t *httmp;
	int i;
	
	for(index = 0, i = 0 ; i < ksize ; i++)
		index += key[i];
	index = index % ht->size;
			
	for(httmp = ht->table[index] ; httmp ; httmp = ht->table[index]->next)
		if(ksize == httmp->ksize)
			if(!memcmp(httmp->key, key, ksize))
				return httmp->value;
	
	return NULL;
}

u_char ritas_ht_hash(u_char *key, u_char ksize) {
	int i;
	u_char index;

	for(index = 0, i = 0 ; i < ksize ; i++)
		index += key[i];
	index = index % ht->size;

	return index;
}
