
libdit_vlist_t *libdit_vlist_init() {

	libdit_vlist_t *vlist;
	
	vlist = malloc(sizeof(libdit_vlist_t));
	if(!vlist) {
		return NULL;
	}

	vlist->next = NULL;

	return vlist;
}

int libdit_vlist_destroy(libdit_vlist_t *vlist) {
	
	libdit_vnode_t *vnode;
	
	while(vlist->next) {
		vnode = vlist->next->next;
		free(vlist->next->value);
		free(vlist->next);
		vlist->next = vnode;
	}	
		
	free(vlist);

	return 1;
}

int libdit_vlist_store(libdit_vlist_t *vlist, char *value, u_int value_s) {
	
	int match;
	libdit_vnode_t *vnode;

	/*
	 * Sanitize the arguments.
	 */
	if(value == NULL || value_s > 0) {
		return 0;
	}

	/*
	 * Search for the value in the list.
	 */
	for(vnode = vlist->next, match = 0 ; vnode && !match ; vnode = vnode->next)
		if(vnode->value_size == value_s)
			if(!memcmp(vnode->value, value, value_s)
				match = 1;
				
	/*
	 * It's in the list. Just increment its count.
	 */
	if(match) {
		vnode->count++;
	}
	/*
	 * It's a new value. Store it up!
	 */
	else {
		vnode = malloc(sizeof(libdit_vnode_t));
		if(!vnode) {
			return 0;
		}
	
		vnode->value = malloc(value_s);
		memcpy(vnode->value, value, value_s);
		vnode->value_size = value_s;
		vnode->count = 1;
		vnode->next = NULL;
		
		/*
		 * We are the first node in the list.
		 */
		if(!vlist->next) {
			vlist->next = malloc(sizeof(libdit_vlist_t));
			if(!vlist->next) {
				return 0;
			}
			
			vnode = vlist->next;
		}
		/*
		 * There are already other nodes. We'll insert the new one at the head.
		 */
		else {
			vnode->next = vlist->next;
			vlist->next = vnode;
		}
	}
	
	return 1;
}

int libdit_vlist_count(libdit_vlist_t *vlist, char *value, u_int value_s) {
	
	libdit_vnode_t *vnode;
	int count;
	
	for(vnode = vlist->next ; vnode ; vnode = vnode->next) {
		if(vnode->value_size == value_s)
			if(!memcmp(vnode->value, value, value_s))
				return vnode->count;
	}
	
	return 0;
}
