#include "peekaboo.h"

int main(int argc, char **argv) {
	
	pcap_t *p;
	char *device;
	u_char *packet;
	struct pcap_pkthdr h;
	struct pcap_stat ps;
	char errbuf[PCAP_ERRBUF_SIZE];
	struct bpf_program filter_code;
	bpf_u_int32 local_net, netmask;
	struct ip *iph;
	struct udphdr *udph;
	ritas_rc_header_t *rch;
	ritas_rb_header_t *rbh;
	ritas_bc_header_t *bch;
	u_char prev;

	if(argc > 1) {
		device = argv[1];
	}
	else {
		device = pcap_lookupdev(errbuf);
		if(device == NULL) {
			fprintf(stderr, "pcap_lookupdev() failed: %s\n", errbuf);
			exit(EXIT_FAILURE);
		}
	}
	printf("Using device: %s\n", device);	
	
	/*
	 * Open the device.
	 */
	p = pcap_open_live(device, SNAPLEN, PROMISC, TIMEOUT, errbuf);
	if(p == NULL) {
		fprintf(stderr, "pcap_open_live() failed: %s\n", errbuf);
		exit(EXIT_FAILURE);
	}
	
	/*
	 * Set the BPF filter.
	 */
	if(pcap_lookupnet(device, &local_net, &netmask, errbuf) == -1) {
		fprintf(stderr, "pcap_lookupnet() failed: %s\n", errbuf);
		pcap_close(p);
		exit(EXIT_FAILURE);
	}
	if(pcap_compile(p, &filter_code, FILTER, 1, netmask) == -1) {
		fprintf(stderr, "pcap_compile() failed: %s\n", pcap_geterr(p));
		pcap_close(p);
		exit(EXIT_FAILURE);
	}
	if(pcap_setfilter(p, &filter_code) == -1) {
		fprintf(stderr, "pcap_setfilter() failed: %s\n", pcap_geterr(p));
		pcap_close(p);
		exit(EXIT_FAILURE);
	}

	/*
	 * Make sure it's Ethernet.
	 */
	 if(pcap_datalink(p) != DLT_EN10MB) {
		fprintf(stderr, "RITAS Peekaboo only works with Ethernet.\n", pcap_geterr(p));
		pcap_close(p);
		exit(EXIT_FAILURE);
	 }
	 	

	/*
	 * Now loop bitch!.
	 */
	for(;;) {
		packet = (u_char *)pcap_next(p, &h);
		if(packet == NULL)
			continue;
		
		printf("==================================\n");
		iph = (struct ip *)(packet + ETHER_ADDR_LEN);
		udph = (struct udphdr *)((u_char *)iph + iph->ip_hl*4);
		rch = (ritas_rc_header_t *)((u_char *)udph + sizeof(*udph));
		rbh = (ritas_rb_header_t *)((u_char *)rch + RITAS_RC_HEADER_SIZE);

		
		printf("[%d] ", rch->pid);

		if(rbh->id.flags & RITAS_ID_AB)
			printf("|ab");
		if(rbh->id.flags & RITAS_ID_VC)
			printf("|vc");
		if(rbh->id.flags & RITAS_ID_MVC)
			printf("|mvc");
		if(rbh->id.flags & RITAS_ID_BC)
			printf("|bc");
		printf("|");
		
		printf("  cid: %d", rbh->id.cid);
		if(rbh->id.flags & RITAS_ID_MVC) 
			printf("  mvcid: %d", rbh->id.mvcid);
		
		printf("\nRB ");
		switch(rbh->type) {
			case RITAS_RB_INITIAL:
				printf("INITIAL");
				break;
			case RITAS_RB_ECHO:
				printf("ECHO");
				break;
			case RITAS_RB_READY:
				printf("READY");
				break;
		}
		printf("  txid: %d", rbh->id.txid);
		printf("  rbid: %d", rbh->id.rbid);
		
		switch(rbh->id.flags) {
			case(RITAS_ID_BC):
				bch = (ritas_bc_header_t *)((u_char *)rbh + RITAS_RB_HEADER_SIZE);
				printf("\nBC");
				printf("  round: %d", bch->round);
				printf("  step: %d", bch->step);
				if(bch->value == 2)
					printf("  value: d");
				else
					printf("  value: %d", bch->value);
				break;
			case(RITAS_ID_MVC):
				break;
			case(RITAS_ID_VC):
				break;
			case(RITAS_ID_AB):
				break;
		}
		printf("\n");
	}
}
