#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "ritas.h"
#include "ritas_rc.h"
#include "ritas_proc.h"
#include "ritas_mbuf.h"
#include "ritas_rb.h"
#include "ritas_vector.h"
#include "ritas_ab.h"
#ifdef MEMWATCH
#include "memwatch.h"
#endif

#define toLL(tv) ((tv.tv_sec*(long long)1000000)+tv.tv_usec)

/*
 * Test suite for RITAS
 */
int main(int argc, char **argv) {
	
	ritas_t *ctx;
	u_short mypid, id;
	int i, n;
	u_char proposal;
	int decision;
	u_char errbuf[RITAS_ERRBUF_SIZE];
	u_int rand;
	struct timeval t0, t1;
	long long diff;
	socklen_t optlen;
	int recvbuf;
	u_char pidbuf[64];
	u_char key[RITAS_KEY_SIZE];
	u_char buf[32];
	u_short buf_s = 32;
	
	if(argc < 2) {
		exit(1);
	}
	
	setlinebuf(stdout);
	mypid = atoi(argv[1]);

	memset(pidbuf, '\0', 64);
	snprintf(pidbuf, 64, "echo %d > pid.%d", getpid(), mypid);
	system(pidbuf);


	ctx = ritas_init(mypid, 4, 1, 4340 + mypid, errbuf); // mypid, n, f
	if(!ctx) {
		printf("%s\n", errbuf);
		exit(1);
	}
	
	memset(key, '\0', RITAS_KEY_SIZE);
	ritas_proc_add_ipv4(ctx, 0, "127.0.0.1", 4340, key);
	ritas_proc_add_ipv4(ctx, 1, "127.0.0.1", 4341, key);
	ritas_proc_add_ipv4(ctx, 2, "127.0.0.1", 4342, key);
	ritas_proc_add_ipv4(ctx, 3, "127.0.0.1", 4343, key);

	sleep(2);
	
	n = ritas_start(ctx);
	if(n < 0) {
		printf("%s\n", ctx->errbuf);
		exit(1);
	}
	
/*
	optlen = sizeof(recvbuf);

	recvbuf = 1024*1024;

	n = setsockopt(ctx->socket, SOL_SOCKET, SO_RCVBUF, &recvbuf, optlen);
	if(n < 0) {
   	    printf("%s(): setsockopt()\n", __func__);
		exit(1);
	}

	n = getsockopt(ctx->socket, SOL_SOCKET, SO_RCVBUF, &recvbuf, &optlen);
	if(n < 0) {
   	    printf("%s(): getsockopt()\n", __func__);
		exit(1);
	}
	printf("SO_RCVBUF: %d bytes\n", recvbuf);
*/	

	sleep(3);

	for(id = 1 ; id <= 10 ; id++) {
		ctx->selt = ctx->st = ctx->rt = 0;
		n = RAND_pseudo_bytes((u_char *)&rand, sizeof(rand));
		if(n == 0) {
			printf("RAND_pseudo_bytes(): error\n");
			exit(1);
		}
		proposal = rand % 2;
		if(mypid == 0)
			sleep(2);
			
		gettimeofday(&t0, NULL);

		if(mypid == 0)
			n = ritas_rb_bcast(ctx, id, (u_char *)&proposal, sizeof(proposal));
		else
			n = ritas_rb_recv(ctx, 0, id, (u_char *)&proposal, sizeof(proposal));

		gettimeofday(&t1, NULL);
		if(n == -1) {
			printf("%s\n", ctx->errbuf);
			exit(1);
		}
		printf("RB t0: %ld.%ld t1: %ld.%ld\n", t0.tv_sec, t0.tv_usec, t1.tv_sec, t1.tv_usec);
		printf("RB rbid %d decided(%d) %d elapsed: %ld us\n\n", id, n, proposal, toLL(t1)-toLL(t0));
		printf("TRB %d....select time: %d us\n", id, ctx->selt);	
		printf("TRB %d...receive time: %d us\n", id, ctx->rt);
		printf("TRB %d......send time: %d us\n", id, ctx->st);		
	}
	


	sleep(3);
	ritas_timestamp();
	printf("[%d] EXIT\n", mypid);
	ritas_destroy(ctx);
	
}
