#ifndef RITAS_RC_H
#define RITAS_RC_H

#include <sys/types.h>
#include <pthread.h>
#include "ritas.h"
#include "ritas_ht.h"
#include "ritas_mbuf.h"
#include "ritas_group.h"

#define RITAS_RC_HASH(rcid)	(rcid % RITAS_RCCB_SIZE)

struct ritas_rccb {
	u_short rcid;				/* reliable channel identification */
#define RITAS_RC_UNACK_SIZE 17	/* should be a prime number */
	ritas_ht_t **unack;
	pthread_mutex_t unack_mutex;
	u_int lastsn;				/* sequence number of last datagram sent */
	//rbcb_t **rbcb;			/* hash table of registered rb control blocks */
	struct ritas_rccb *prev;
	struct ritas_rccb *next;
};
typedef struct ritas_rccb ritas_rccb_t;


struct ritas_rc_header {
	u_short type;
#define RITAS_RC_DATA	0x01
#define RITAS_RC_ACK	0x02
	u_short rcid;
	union {
		u_int sn;
		u_int ack;
	} ht;
}; 
typedef struct ritas_rc_header ritas_rc_header_t;
#define sn	ht.sn
#define ack	ht.ack


ritas_rccb_t *ritas_rc_init(ritas_t *ctx, u_short rcid);
void ritas_rc_destroy(ritas_t *ctx, ritas_rccb_t *rccb);
int ritas_rc_output(ritas_t *ctx, ritas_rccb_t *rccb, ritas_mbuf_t *m);
void ritas_rc_input(ritas_t *ctx, ritas_mbuf_t *m, ritas_proc_t *proc);



#endif /* RITAS_RC_H */
