
struct ritas_rc_procmbuf *ritas_rc_unack_lookup(struct ritas_rc_procmbuf *pmtable, ritas_proc_t *proc, u_short sn) {
	
	int i;
	struct ritas_rc_procmbuf *pm;

	i = RITAS_RC_UNACK_HASH(proc->addr, sn);
	
	for(pm = pmtable[i] ; pm ; pm = pm->next)
		if(pm->proc
		|| ((ritas_rc_header_t *)pm->mbuf->head)->sn == sn)
			return pm->proc;
			
	return NULL;
}

int ritas_rc_unack_remove(struct ritas_rc_procmbuf *pm) {
	
	pm->prev->next = pm->next;
	pm->next->prev = pm->prev;
	
	pm->mbuf->usecount--;
	if(pm->mbuf->usecount <= 0)
		ritas_mbuf_destroy(mbuf);
	
	free(pm);
	
	return 1;
}


int ritas_rc_unack_add(struct ritas_rc_procmbuf *pmtable, ritas_mbuf_t *mbuf, ritas_proc_t *proc) {

	int i;
	struct ritas_rc_procmbuf *pmtmp;

	i = RITAS_RC_UNACK_HASH(proc->addr, ((ritas_rc_header_t *)mbuf->head)->sn);
	
	pmtmp = pmtable[i];
	pmtable[i] = (struct ritas_rc_procmbuf *)malloc(sizeof(struct ritas_rc_procmbuf));
	pmtable[i]->prev = NULL;
	pmtable[i]->next = pmtmp;
	pmtmp->prev = pmtable[i];	
	
	pmtable[i]->nrexmt = 0;
	pmtable[i]->rto = proc->rttinfo.rto;
	pmtable[i]->mbuf = mbuf;
	pmtable[i]->proc = proc;
	
	return 1;
}


void ritas_rc_rtt_timer(ritas_t *ctx) {

	ritas_rccb_t *rccb;
	ritas_rc_rttproc *rttproc;
	int i, n;

	for(;;) {
		sleep(RITAS_RC_RRT_TIMER);
		// go through unack packets decrease the timers
		// and if any reaches zero resend the packet
		// and update the RTT information.
		for(i = 0 ; i < RITAS_RCCB_SIZE ; i++) {
			for(rccb = ctx->rccb[i] ; rccb ; rccb = rccb->next) {
				for(j = 0 ; j < RITAS_RC_UNACK_SIZE ; j++) {
					for(pm = rccb->unack[j] ; pm ; pm = pm->next) {
						pm->rto -= 1.0;
						/*
						 * Timeout?
						 */
						if(pm->rto <= 0.0) {
							/*
							 * Retransmit
							 */
							n = ritas_send(ctx, pm->proc, pm->mbuf);
							if(n < 0) {
								// we have to exit cleanly
							}
							pm->nrexmt++;	/* update # retransmisssions */
							/*
							 * Have we retransmited enough?
							 */
							if(pm->nrexmt > RTT_MAXNREXMT) {
								// remove mbuf
								ritas_rc_unack_remove(pm);
							}
							else
								// double RTO
								RITAS_RC_RTO_DOUBLE(pm->proc.rttinfo);
						}
					}
				}
			}
		}
	}
}

static float ritas_rc_rtt_minmax(float rto) {
	if (rto < RTT_RXTMIN)
		rto = RTT_RXTMIN;
	else if (rto > RTT_RXTMAX)
		rto = RTT_RXTMAX;
	return (rto);
}


void ritas_rc_rtt_init(struct ritas_rc_rtt *ptr) {
	struct timeval tv;

	gettimeofday(&tv, NULL);
	ptr->rtt_base = tv.tv_sec;   /* # sec since 1/1/1970 at start */

	ptr->rtt_rtt = 0;
	ptr->rtt_srtt = 0;
	ptr->rtt_rttvar = 0.75;
	ptr->rtt_rto = rtt_minmax(RTT_RTOCALC(ptr));
	/* first RTO at (srtt + (4 * rttvar)) = 3 seconds */
}

