#ifndef RITAS_RC_RTT_H
#define RITAS_RC_RTT_H

#include "ritas_mbuf.h"
#include "ritas_group.h"

#define RITAS_RC_RTT_TIMER 1	/* define our timer to be 1 seconds */

/*
 * Peer specific RTT information
 */
struct ritas_rc_rtt {
	float rtt;		/* most recent measured RTT, in seconds */
	float srtt;		/* smoothed RTT estimator, in seconds */
	float rttvar;	/* smoothed mean deviation, in seconds */
	float rto;		/* current RTO to use, in seconds */
};
#define RTT_RXTMIN		2
#define RTT_RXTMAX		60
#define RTT_MAXNREXMT	3



void ritas_rc_rtt_timer(int sig_num);

/*
 * Calculate the RTO value based on current estimators:
 *      smoothed RTT plus four times the deviation
 */
#define RITAS_RC_RTOCALC(ptr)	((ptr)->srtt + (4.0 * (ptr)->rttvar))

/*
 * This is called the first time any packet is sent. The measured RTT is
 * set to 0 and the RTT and mean deviation estimators are set to 0 and 0.75,
 * respectively, giving an initial RTO of 3 seconds.
 */
void ritas_rc_rtt_init(struct ritas_rc_rtt *ptr);

/*
 * Sets the retransmission counter to 0. This funcion should be called
 * whenever a new packet is sent for the first time.
 */
void ritas_rc_rtt_newpack(struct ritas_rc_rtt *ptr);

/*
 * Returns the current RTO in seconds. The return value can then be used
 * as the argument to alarm().
 */
void ritas_rc_rtt_start(struct ritas_rc_rtt *ptr);

/*
 * This is called after an ack is received to update the RTT estimators
 * and calculate the new RTO.
 */
void ritas_rc_rtt_stop(struct ritas_rc_rtt *ptr, u_int ms);

/*
 * This is called when the retransmission timeout expires.
 * Applies the exponencial backoff.
 */
void ritas_rc_rtt_timeout(struct ritas_rc_rtt *ptr);

/*
 * Returns the current timestamp for the caller to store as an unsigned
 * 32-bit integer in the datagram being sent.
 */
u_int ritas_rc_rtt_ts(struct ritas_rc_rtt *ptr);

/*
 * Makes certain that the RTO is between the upper and lower limits
 */
static float ritas_rc_rtt_minmax(float rto); 

#endif	/* RITAS_RC_RRT_H */
