#ifndef RITAS_H
#define RITAS_H

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <pthread.h>
#include <openssl/evp.h>
//#include "ritas_aproxy.h"
#include "ritas_ht.h"
#include "ritas_list.h"
#ifdef MEMWATCH
#include "memwatch.h"
#endif

#ifdef DEBUG
#define dprintf(...) printf(__VA_ARGS__)
#else
#define dprintf(...)
#endif

#ifdef AB_DEBUG
#define abprintf(...) printf(__VA_ARGS__)
#else
#define abprintf(...)
#endif

#ifdef RB_DEBUG
#define rbprintf(...) printf(__VA_ARGS__)
#else
#define rbprintf(...)
#endif

#ifdef BC_DEBUG
#define bcprintf(...) printf(__VA_ARGS__)
#else
#define bcprintf(...)
#endif

#ifdef MVC_DEBUG
#define mvcprintf(...) printf(__VA_ARGS__)
#else
#define mvcprintf(...)
#endif

#ifdef VC_DEBUG
#define vcprintf(...) printf(__VA_ARGS__)
#else
#define vcprintf(...)
#endif


#define RITAS_BC	1
#define RITAS_MVC	2
#define RITAS_VC	3
#define RITAS_AB	4

#define toLL(tv) ((tv.tv_sec*(long long)1000000)+tv.tv_usec)

/*
 * The context control structure over ritas's operation.
 * Contains all the necessary data to maintain state on an intrusion-tolerant 
 * communication. This structure is completely opaque to the programmer via the
 * ritas_t datatype, its internal contents should never be needed to be 
 * directly accessed.
 */
struct ritas_context {
#define RITAS_ERRBUF_SIZE 64	
	char errbuf[RITAS_ERRBUF_SIZE];
	int socket;
	struct ritas_process **group;
	u_short n;
	u_short f;
	u_short pid;
	struct ritas_abcb *abcb;
#define RITAS_HT_RCCB_SIZE 251	// should be a prime number	
	ritas_ht_t *ht_rccb;
#define RITAS_HT_OOC_SIZE 251
	ritas_ht_t *ht_ooc;
	ritas_list_t *l_ooc;
	struct ritas_aproxy *aproxy;
	u_char error;
	pthread_mutex_t err_mutex;
	pthread_t scheduler_net_thread;
	pthread_t scheduler_thread;
	ritas_list_t *scheduler_l;
	pthread_mutex_t scheduler_mutex;
	pthread_cond_t scheduler_cond;
	int scheduler_v;

	struct timeval t_init;

	struct timeval t0;
	int rt, st, selt;
	pthread_mutex_t temp_mutex;

	pthread_mutex_t pr_mutex;
	pthread_cond_t pr_cond;
	u_short pr_count;

	/*
	 * RBTO
	 */
	pthread_mutex_t rbto_mutex;
	ritas_list_t *rbto_l; // List for all RBTOs
	long long srtt;
	long long rttvar;
	long long rto;
	struct itimerval itimer;
	ritas_ht_t *rc_rbto_ht; // HT for RC RBTOs	
};	
typedef struct ritas_context ritas_t;


ritas_t *ritas_init(u_short pid, u_short n, u_short f, u_short port, char *errbuf);
void ritas_destroy(ritas_t *ctx);
void ritas_renew(ritas_t *ctx);
int ritas_start(ritas_t *ctx);
void ritas_scheduler_net(ritas_t *ctx);
void ritas_scheduler(ritas_t *ctx);
void ritas_timestamp();

#ifdef TCP
void ritas_accept(ritas_t *ctx);
void ritas_connect(ritas_t *ctx);
void ritas_net(ritas_t *ctx);
#endif

#endif /* RITAS_H */
