#include <stdlib.h>
#include "ritas_ab_aht.h"

ritas_ab_aht_t *ritas_ab_aht_init() {
	
	ritas_ab_aht_t *aht;
	
	aht = malloc(sizeof(*aht));
	if(!aht)
		goto bad;
		
	aht->ht = ritas_ht_init(RITAS_AB_AHT_HT_SIZE);
	if(!aht->ht)
		goto bad;
		
	aht->l = ritas_list_init();
	if(!aht->l)
		goto bad;
		
	return aht;
	
bad:
	if(aht) {
		ritas_ab_aht_destroy(aht);
	}
	return NULL;
}
	
void ritas_ab_aht_destroy(ritas_ab_aht_t *aht) {
	
	if(!aht)
		return;
		
	ritas_list_destroy(aht->l);
	ritas_ht_destroy(aht->ht, (void(*)(void *))ritas_ab_aht_list_entry_destroy);
	free(aht);
}

void ritas_ab_aht_renew(ritas_ab_aht_t *aht) {
	
	ritas_ht_renew(aht->ht, NULL);
	aht->l->first = aht->l->last = NULL;
}

ritas_ab_aht_list_entry_t *ritas_ab_aht_list_entry_init() {
	
	ritas_ab_aht_list_entry_t *le;
	
	le = malloc(sizeof(*le));
	if(!le)
		return NULL;
	
	le->prev = le->next = NULL;
	le->m = NULL;
	
	return le;
}
	
void ritas_ab_aht_list_entry_destroy(ritas_ab_aht_list_entry_t *le) {
	ritas_mbuf_destroy(le->m);
	free(le);
}

int ritas_ab_aht_insert(ritas_ab_aht_t *aht, u_char *hash, u_short hash_s, ritas_mbuf_t *m) {
	
	int n;
	ritas_ab_aht_list_entry_t *le;

	le = ritas_ab_aht_list_entry_init();
	if(!le)
		return -1;
	le->m = m;
	
	n = ritas_list_enqueue(aht->l, le);
	if(n == -1)
		return -1;

	n = ritas_ht_add(aht->ht, hash, hash_s, le);
	if(n == -1)
		return -1;
		
	return 1;	
}

void ritas_ab_aht_update(ritas_ab_aht_entry_t *e, ritas_mbuf_t *m) {

	ritas_ab_aht_list_entry_t *le;
		
	le = e->value;
	le->m = m;
	
}

void *ritas_ab_aht_remove(ritas_ab_aht_t *aht, u_char *hash, u_short hash_s) {
	
	ritas_ab_aht_list_entry_t *le;
	void *m;
	
	le = ritas_ht_remove(aht->ht, hash, hash_s);
	if(!le)
		return NULL;
	
	ritas_list_remove(aht->l, le);
	
	m = le->m;
	free(le);
	
	return m;
}

ritas_ab_aht_entry_t *ritas_ab_aht_lookup(ritas_ab_aht_t *aht, u_char *hash, u_short hash_s) {
	
	return ritas_ht_lookup_ext(aht->ht, hash, hash_s);
}		

ritas_list_t *ritas_ab_aht_list(ritas_ab_aht_t *aht) {
	
	ritas_list_t *l;
	ritas_ab_aht_list_entry_t *le;
	int n;
	
	l = ritas_list_init();
	if(!l)
		return NULL;
	
	for(le = (ritas_ab_aht_list_entry_t *)aht->l->first ; le ; le = le->next) {
		n = ritas_list_enqueue(l, le->m);
		if(n == -1) {
			free(l);
			return NULL;
		}
	}
	return l;
}
