#include "ritas_bc_abba.h"
#include "ritas_thsig.h"
#include "ritas_thcoin.h"
#include <string.h>
#include <openssl/rand.h>

int ritas_bc_input(ritas_t *ctx, ritas_bccb_t *bccb, ritas_mbuf_t *m) {
	
	ritas_bc_abba_header_t *bch;
	ritas_bc_abba_coin_header_t *bch_coin;
	int i, j, valid, sigsize, r;
	u_char b, value, *just, *p;
	u_short type, round, just_s, step;
	u_int coin_name;
	ritas_mbuf_t *mout;
	ritas_bc_abba_prevote_just_t *pvjp_aux;

	step = *(u_short *)m->head;
	// this is for protocol corrrectness debugging purposes!!
	/*
	if(m->proc->id == bccb->id && *(u_short *)m->head != RITAS_BC_ABBA_PREVOTE) {
		ritas_mbuf_destroy(m);
		return 1;
	}
	*/
	
	/*
	 * No, I am not ashamed of using goto!
	 */
	if(step == RITAS_BC_ABBA_PREVOTE && bccb->step != 2)
		goto step_2;
	
	
	switch(bccb->step) {
		/**********************
		 ****** Step 4. *******
		 ********************** 
		 */
		case 4:
		/*
		 * Sanitize message. Check for if the current type and round correspond.
		 */
		if(*(u_short *)m->head != RITAS_BC_ABBA_COIN) {
			ritas_mbuf_destroy(m);
			return 1;
		}

		bch_coin = (ritas_bc_abba_coin_header_t *)m->head;
		dprintf("[%d] BC in COIN for round %d from %d\n", bccb->id, bch_coin->round, m->proc->id);
	
		if(bch_coin->round != bccb->round) {
			ritas_mbuf_destroy(m);
			return 1;
		}

		/*
		 * Verify coin share and increment the count if valid.
		 */
		valid = ritas_thcoin_verify_share(bccb->thcoin, bch_coin->share, m->proc->id);
		if(!valid) {
			dprintf("[%d] BC the coin share is invalid!\n", bccb->id);
			ritas_mbuf_destroy(m);
			return 1;
		}
		ritas_thcoin_add_share(bccb->thcoin, bch_coin->share, m->proc->id);			
		bccb->coin_share_count++;
		
		/*
		 * When the received coin shares reach n-f, combine them into F(TID, r)
		 * and proceed to the next round.
		 */
		if(bccb->coin_share_count == ctx->n - ctx->f) {
			bccb->coin_share_count = 0;
			bccb->round++;
			bccb->step = 1;
			ritas_rc_destroy(ctx, bccb->rccb);
			bccb->rccb = NULL;
		
			dprintf("[%d] BC ********** Round %d ***********\n", bccb->id, bccb->round);
			
			/*
			 * Advance the pre-vote justification pointers. Kinky stuff.
			 */
			for(i = 0 ; i < 2 ; i++) {
				free(bccb->pvjp_rm1->data[i]);
				memset(bccb->pvjp_rm1, '\0', sizeof(*bccb->pvjp_rm1));
			}
			pvjp_aux = bccb->pvjp_rm1;
			bccb->pvjp_rm1 = bccb->pvjp;
			bccb->pvjp = pvjp_aux;
			

			bccb->coin_flip = ritas_thcoin_combine_shares(bccb->thcoin);
			dprintf("[%d] BC coin flip result is %d\n", bccb->id, bccb->coin_flip);		
		}
		else {
			ritas_mbuf_destroy(m);
			return 1;
		}

		/**********************
		 ****** Step 1. *******
		 ********************** 
		 */
		case 1:
		bch = (ritas_bc_abba_header_t *)m->head;
		/*
		 * When in round r = 1, we're handling pre-process messages. Verify the
		 * signature and update the count for the value received.
		 */
		if(bccb->round == 1) {
			dprintf("[%d] BC in PRE-PROCESS value %d from %d\n", bccb->id, bch->value, m->proc->id);
			valid = ritas_bc_abba_verify_share(bccb, bch->share, bch->type, bch->round, bch->value, m->proc->id);
			if(!valid) {
				dprintf("[%d] BC the signature share is invalid!\n", bccb->id);
				ritas_mbuf_destroy(m);
				return 1;
			}
			
			bccb->vcount[RITAS_BC_TOTAL]++;
			bccb->vcount[!bch->value ? RITAS_BC_ZERO : RITAS_BC_ONE]++;
			bccb->preprocess_share[bch->value][m->proc->id] = malloc(RITAS_THSIG_SHARE_SIZE);
			if(!bccb->preprocess_share[bch->value][m->proc->id]) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): error in malloc	\n", __func__);
				return -1;
			}									
			memcpy(bccb->preprocess_share[bch->value][m->proc->id], bch->share, RITAS_THSIG_SHARE_SIZE);			
		}
		
		/*
		 * When in round r = 1, this condition returns true when the number of
		 * received pre-process messages reaches n-f.
		 * When in round r > 1, this condition always returns true.
		 */
prevote:
		valid = 0;
		if(bccb->vcount[RITAS_BC_TOTAL] == ctx->n - ctx->f)
			if(bccb->round == 1)
				valid = 1;
		 	else // round > 1
		 		if(bccb->vcount[RITAS_BC_DEFAULT] == ctx->n - ctx->f)
		 			valid = 1;
		 		else if(bccb->prevote_share_count[0] >= ctx->n - ctx->f 
		 		     || bccb->prevote_share_count[1] >= ctx->n - ctx->f)
		 			valid = 1;
		 		
		if(valid) {
			bccb->step = 2;
			r = ritas_rc_renew(ctx, bccb->rccb_pre, bccb);
			if(r < 0) {
				return -1;
			}
			bccb->prevote_share_count[0] = bccb->prevote_share_count[1] = 0;
			
			bch->type = RITAS_BC_ABBA_PREVOTE;
			bch->round = bccb->round;

			/*
			 * In round r = 1, our vote b is just the majority value of
			 * pre-process messages.
			 */
			ritas_thsig_reset(bccb->thsig, 0);
			if(bccb->round == 1) {
				bch->subtype = RITAS_BC_ABBA_PREVOTE_ROUND_ONE;				
				b = bccb->vcount[RITAS_BC_ONE] >= bccb->vcount[RITAS_BC_ZERO] ? 1 : 0;			
		
				/*
				 * Produce justification. Collect f+1 pre-process votes for value
				 * b, and use their shares to generate a threshold signature.
				 */
				for(i = 0 ; i < ctx->n ; i++) {
					if(bccb->preprocess_share[b][i]) {
						ritas_thsig_add_share(bccb->thsig, bccb->preprocess_share[b][i], i);
						if(bccb->thsig->share_count == ctx->f+1)
							break;
					}
				}
				sigsize = ritas_thsig_combine_shares(bccb->thsig, ctx->f+1, (u_char *)bch + RITAS_BC_ABBA_HEADER_SIZE);		
				if(sigsize < 0) {
					printf("BC error in share combination!!\n");
					exit(1);
				}
			}

			/*
			 * In round r > 1, our vote b is either the last coin flip, in case
			 * all main-votes were abstain, or is 1 or 0, in case there was a
			 * main vote for 1 or 0, respectivelly.
			 */
			else {
				/*
				 * In case our pre-vote is soft, the justification is a
				 * threshold signature on the message 
				 * (TID,main-vote,r-1,abstain).
				 */
				if(bccb->vcount[RITAS_BC_DEFAULT] == bccb->vcount[RITAS_BC_TOTAL]) {
					dprintf("[%d] BC to cast a soft pre-vote on %d\n", bccb->id, bccb->coin_flip);
					bch->subtype = RITAS_BC_ABBA_PREVOTE_SOFT;					
					b = bccb->coin_flip;
					for(i = 0 ; i < ctx->n ; i++) {
						if(bccb->mainvote_share[RITAS_BC_DEFAULT][i]) {
							ritas_thsig_add_share(bccb->thsig, bccb->mainvote_share[RITAS_BC_DEFAULT][i], i);
							if(bccb->thsig->share_count == ctx->n - ctx->f)
								break;
						}
					}
					sigsize = ritas_thsig_combine_shares(bccb->thsig, ctx->n - ctx->f, (u_char *)bch + RITAS_BC_ABBA_HEADER_SIZE);
					if(sigsize < 0) {
						printf("BC error in share combination!!\n");
						exit(1);
					}
				}
				/*
				 * Otherwise, our pre-vote is hard. In that case the 
				 * justification is a threshold signature on the message
				 * (TID,pre-vote,r-1,b).
				 */
				else {
					bch->subtype = RITAS_BC_ABBA_PREVOTE_HARD;					
					b = bccb->vcount[RITAS_BC_ONE] ? 1 : 0;
					dprintf("[%d] BC to cast a hard pre-vote on %d\n", bccb->id, b);
					for(i = 0 ; i < ctx->n ; i++) {
						if(bccb->prevote_share[b][i])  {
							ritas_thsig_add_share(bccb->thsig, bccb->prevote_share[b][i], i);
							if(bccb->thsig->share_count == ctx->n - ctx->f)
								break;
						}
					}
					sigsize = ritas_thsig_combine_shares(bccb->thsig, ctx->n - ctx->f, (u_char *)bch + RITAS_BC_ABBA_HEADER_SIZE);
					if(sigsize < 0) {
						printf("BC error in share combination!!\n");
						exit(1);
					}
				}
			}

			m->tail = (u_char *)bch + RITAS_BC_ABBA_HEADER_SIZE + sigsize;
			if(m->tail > m->head + RITAS_MBUF_DATA_SIZE) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
				return -1;
			}						
			
			bch->value = b;
			#ifdef BC_EVIL
			if(ctx->pid < ctx->f)
				if(bch->subtype == RITAS_BC_ABBA_PREVOTE_HARD)
					bch->value = (bch->value & 0x01 ? 0 : 1);
			#endif
			
	
			/*
			 * Produce an S-signature share on the message: (TID,pre-vote,r,b).
			 */
			ritas_bc_abba_share(bccb, bch->share, bch->type, bch->round, bch->value);

			/*
			 * Broadcast the pre-vote message.
			 */
			ritas_rc_broadcast(ctx, bccb->rccb_pre, m);	
			for(i = 0 ; i < 2 ; i++) {
				for(j = 0 ; j < ctx->n ; j++) {
					free(bccb->prevote_share[i][j]);
					bccb->prevote_share[i][j] = NULL;
				}
			}
			memset(bccb->vcount, '\0', 4*sizeof(u_short));
		}
		else {
			ritas_mbuf_destroy(m);
			return 1;
		}

		break;
		
		/**********************
		 ****** Step 2. *******
		 ********************** 
		 */
		case 2:
step_2:		
		bch = (ritas_bc_abba_header_t *)m->head;
		dprintf("[%d] BC in PRE-VOTE value %d from %d\n", bccb->id, bch->value, m->proc->id);
		just = m->head + RITAS_BC_ABBA_HEADER_SIZE;
		just_s = m->tail - m->head - RITAS_BC_ABBA_HEADER_SIZE;

		/*
		 * Verify the signature share.
		 */
		valid = ritas_bc_abba_verify_share(bccb, bch->share, bch->type, bch->round, bch->value, m->proc->id);
		if(!valid) {
			dprintf("[%d] BC the signature share is invalid!\n", bccb->id);
			return 1;
		}
		
		if(!bccb->prevote_share[bch->value][m->proc->id]) {
			bccb->prevote_share[bch->value][m->proc->id] = malloc(RITAS_THSIG_SHARE_SIZE);
			memcpy(bccb->prevote_share[bch->value][m->proc->id], bch->share, RITAS_THSIG_SHARE_SIZE);
		}
		
		/*
		 * Verify the justification.
		 */
		if(bccb->round == 1) {
			valid = ritas_bc_abba_verify(bccb, just, just_s, ctx->f+1, RITAS_BC_ABBA_PREPROCESS, 1, bch->value);
		} 
		else {
			if(bch->subtype == RITAS_BC_ABBA_PREVOTE_HARD) {
				type = RITAS_BC_ABBA_PREVOTE;
				round = bch->round - 1;
				value = bch->value;
			}
			else if(bch->subtype == RITAS_BC_ABBA_PREVOTE_SOFT) {
				type = RITAS_BC_ABBA_MAINVOTE;
				round = bch->round - 1;
				value = RITAS_BC_DEFAULT;
			}
			valid = ritas_bc_abba_verify(bccb, just, just_s, ctx->n - ctx->f, type, round, value);
		}
		if(!valid) {
			dprintf("[%d] BC the justification is invalid!\n", bccb->id);
			ritas_mbuf_destroy(m);
			return 1;
		}
	
		/*
		 * Save the justification. It may need to be reused to cast an abstain
		 * mainvote.
		 */
		if(!bccb->pvjp->data[bch->value]) {
			bccb->pvjp->type[bch->value] = bch->subtype;
			bccb->pvjp->size[bch->value] = just_s;
			bccb->pvjp->data[bch->value] = malloc(just_s);
			memcpy(bccb->pvjp->data[bch->value], just, just_s);
		}
		 

		/*
		 * Shameless voodoo.
		 */
		bccb->prevote_share_count[bch->value]++;
		if(bccb->step == 1)// && bccb->vcount[RITAS_BC_TOTAL] == ctx->n - ctx->f)
			goto prevote;
		if(bccb->step != 2) {
			ritas_mbuf_destroy(m);
			return 1;
		}

		/*
		 * Increment the value count.
		 */
		bccb->vcount[RITAS_BC_TOTAL]++;
		bccb->vcount[!bch->value ? RITAS_BC_ZERO : RITAS_BC_ONE]++;
		
		/*
		 * Wait until we collect n-f pre-vote messages.
		 */
		if(bccb->vcount[RITAS_BC_TOTAL] == ctx->n - ctx->f) {
			bccb->step = 3;
			bccb->rccb = ritas_rc_init(ctx, bccb, RITAS_RC_PROTO_BC);
			if(!bccb->rccb) {
				return -1;
			}

			bch->type = RITAS_BC_ABBA_MAINVOTE;
			bch->subtype = 0;
			bch->round = bccb->round;

			/*
			 * In case we receive conflicting votes, our main-vote is abstain.
			 */
			if(bccb->vcount[RITAS_BC_ONE] > 0 && bccb->vcount[RITAS_BC_ZERO] > 0) {
				dprintf("[%d] BC to cast a main-vote on ABSTAIN\n", bccb->id);
				b = RITAS_BC_DEFAULT;
				/*
				 * Justification is just two conflicting pre-vote justifications.
				 */		 
				p = (u_char *)bch + RITAS_BC_ABBA_HEADER_SIZE;
				pvjp_aux = (bccb->round >= 1 ? bccb->pvjp : bccb->pvjp_rm1);
				
				for(i = 0 ; i < 2 ; i++) {
					*(u_short *)p = pvjp_aux->type[i];
					p += sizeof(pvjp_aux->type[i]);
					*(u_short *)p = pvjp_aux->size[i];
					p += sizeof(pvjp_aux->size[i]);
					memcpy(p, pvjp_aux->data[i], pvjp_aux->size[i]);
					p += pvjp_aux->size[i];
					
					valid = ritas_bc_abba_verify(bccb, pvjp_aux->data[i], pvjp_aux->size[i], ctx->n - ctx->f, RITAS_BC_ABBA_MAINVOTE, bccb->round-1, RITAS_BC_DEFAULT);
				}
				
				/*
				 * Adjust mbuf
				 */
				m->tail = p;
				if(m->tail > m->head + RITAS_MBUF_DATA_SIZE) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
					return -1;
				}
			}
			
			/*
			 * If all n-f votes are for either 1 or 0, we main-vote accordingly.
			 */
			else {
				b = bccb->vcount[RITAS_BC_ONE] ? 1 : 0;		
				dprintf("[%d] BC to cast a main-vote on %d\n", bccb->id, b);

				/*
				 * Produce justification.
				 */
				ritas_thsig_reset(bccb->thsig, 0);
				for(i = 0 ; i < ctx->n ; i++) {
					if(bccb->prevote_share[b][i])  {
						ritas_thsig_add_share(bccb->thsig, bccb->prevote_share[b][i], i);
						if(bccb->thsig->share_count == ctx->n - ctx->f)
							break;
					}
				}
				sigsize = ritas_thsig_combine_shares(bccb->thsig, ctx->n - ctx->f, (u_char *)bch + RITAS_BC_ABBA_HEADER_SIZE);
				if(sigsize < 0) {
					printf("BC error in share combination!!\n");
					exit(1);
				}				

				/* 
				 * Adjust mbuf
				 */
				m->tail = (u_char *)bch + RITAS_BC_ABBA_HEADER_SIZE + sigsize;
				if(m->tail > m->head + RITAS_MBUF_DATA_SIZE) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
					return -1;
				}
			}		
			
			bch->value = b;
			#ifdef BC_EVIL
			if(ctx->pid < ctx->f)
				if(bch->value == RITAS_BC_ONE)
					bch->value = RITAS_BC_ZERO;
				else if(bch->value == RITAS_BC_ZERO)
					bch->value = RITAS_BC_ONE;
			#endif
			
			/*
			 * Produce threshold S-signature on the message: (TID,main-vote,r,b).
			 */
			ritas_bc_abba_share(bccb, bch->share, bch->type, bch->round, bch->value);

			/*
			 * Broadcast the main-vote message.
			 */
			ritas_rc_broadcast(ctx, bccb->rccb, m);

			for(i = 0 ; i < 3 ; i++) {
				for(j = 0 ; j < ctx->n ; j++) {
					free(bccb->mainvote_share[i][j]);
					bccb->mainvote_share[i][j] = NULL;
				}
			}

			/*
			 * If we have a decision (from the previous round) then stop.
			 */
			if(bccb->decision == 1 || bccb->decision == 0) {
				if(!bccb->upper) {
					//ritas_bc_destroy(ctx, bccb);
					pthread_mutex_lock(&ctx->pr_mutex);
					ctx->pr_count--;
					pthread_cond_broadcast(&ctx->pr_cond);					
					pthread_mutex_unlock(&ctx->pr_mutex);				
				}
				return 1;	
			}	
					
			memset(bccb->vcount, '\0', 4*sizeof(u_short));
		}
		else {
			ritas_mbuf_destroy(m);
			return 1;
		}
		
		break;

		/**********************
		 ****** Step 3. *******
		 ********************** 
		 */		 
		case 3:
		bch = (ritas_bc_abba_header_t *)m->head;
		if(bch->value == RITAS_BC_DEFAULT)
			dprintf("[%d] BC in MAIN-VOTE value ABSTAIN from %d\n", bccb->id, m->proc->id);
		else
			dprintf("[%d] BC in MAIN-VOTE value %d from %d\n", bccb->id, bch->value, m->proc->id);
	
		/*
		 * Verify the signature share.
		 */
		valid = ritas_bc_abba_verify_share(bccb, bch->share, bch->type, bch->round, bch->value, m->proc->id);
		if(!valid) {
			dprintf("[%d] BC the signature share is invalid\n", bccb->id);
			return 1;
		}

		if(!bccb->mainvote_share[bch->value][m->proc->id]) {
			bccb->mainvote_share[bch->value][m->proc->id] = malloc(RITAS_THSIG_SHARE_SIZE);
			memcpy(bccb->mainvote_share[bch->value][m->proc->id], bch->share, RITAS_THSIG_SHARE_SIZE);
		}

		/*
		 * Verify the justification.
		 */
		if(bch->value == RITAS_BC_DEFAULT) {
			p = m->head + RITAS_BC_ABBA_HEADER_SIZE;
			
			for(valid = 0, value = 0 ; value < 2 ; value++) {
				type = *(u_short *)p;
				p += sizeof(type);
				just_s = *(u_short *)p;
				just = p += sizeof(just_s);
				p += just_s;
				
				if(type == RITAS_BC_ABBA_PREVOTE_ROUND_ONE) 
					valid += ritas_bc_abba_verify(bccb, just, just_s, ctx->f+1, RITAS_BC_ABBA_PREPROCESS, 1, value);
				else if(type == RITAS_BC_ABBA_PREVOTE_SOFT) {
					valid += ritas_bc_abba_verify(bccb, just, just_s, ctx->n - ctx->f, RITAS_BC_ABBA_MAINVOTE, bccb->round-1, RITAS_BC_DEFAULT);
				}
				else {
					valid += ritas_bc_abba_verify(bccb, just, just_s, ctx->n - ctx->f, RITAS_BC_ABBA_PREVOTE, bccb->round-1, value);
				}			
			}
			if(valid < 2)
				valid = 0;
		}
		else {
			just = m->head + RITAS_BC_ABBA_HEADER_SIZE;
			just_s = m->tail - m->head - RITAS_BC_ABBA_HEADER_SIZE;
			valid = ritas_bc_abba_verify(bccb, just, just_s, ctx->n - ctx->f, RITAS_BC_ABBA_PREVOTE, bch->round, bch->value);
		}
		if(!valid) {
			dprintf("[%d] BC the justification is invalid\n", bccb->id);
			return 1;
		}

		/*
		 * Increment the value count.
		 */
		bccb->vcount[RITAS_BC_TOTAL]++;
		bccb->vcount[bch->value]++;

		/*
		 * Wait for n-f main-vote messages.
		 */
		if(bccb->vcount[RITAS_BC_TOTAL] == ctx->n - ctx->f) {
			bccb->step = 4;
			r = ritas_rc_renew(ctx, bccb->rccb, bccb);
			if(r < 0) {
				return -1;
			}
			
			/*
			 * If all main-votes are for either 1 or 0, we decide accordingly
			 * and continue for just one more round. Otherwise, simply proceed.
			 */
			if(bccb->vcount[RITAS_BC_ONE] == ctx->n - ctx->f)
				bccb->decision = 1;
			else if(bccb->vcount[RITAS_BC_ZERO] == ctx->n - ctx->f)
				bccb->decision = 0;

			if(bccb->decision != RITAS_BC_DEFAULT) {
				dprintf("[%d] BC we have a decision of %d! (must continue...)\n", bccb->id, bccb->decision);
				ritas_register_decision_time(ctx->pid, bccb->id, bccb->decision, bccb->round, ctx->t0);
				
				mout = ritas_mbuf_init();
				mout->head = mout->tail - sizeof(bccb->decision);
				*mout->head = bccb->decision;

				if(!bccb->upper) {
					/*
					 * Deliver to the application
					 */
					//ctx->aproxy->mvalue = mout; 
					//pthread_mutex_lock(&ctx->aproxy->mutex);
					//ctx->aproxy->count = 1;
					//pthread_cond_broadcast(&ctx->aproxy->cond);
					//pthread_mutex_unlock(&ctx->aproxy->mutex);
				}
				else {
				}
			}		
		
			/*
			 * Generate a coin share for the coin (TID,r) and broadcast it
			 * in a coin message.
			 */
			bch_coin = (ritas_bc_abba_coin_header_t *)m->head;
			m->tail = m->head + RITAS_BC_ABBA_COIN_HEADER_SIZE;
			bch_coin->type = RITAS_BC_ABBA_COIN;
			bch_coin->round = bccb->round;
			
			coin_name = bccb->id << sizeof(bccb->id)*8;
			coin_name += bccb->round;
			ritas_thcoin_reset(bccb->thcoin, (u_char *)&coin_name, sizeof(coin_name));
			ritas_thcoin_generate_share(bccb->thcoin, bch_coin->share);

			//valid = ritas_thcoin_verify_share(bccb->thcoin, bch_coin->share, ctx->pid);

			ritas_rc_broadcast(ctx, bccb->rccb, m);
		}
		else {
			ritas_mbuf_destroy(m);
			return 1;
		}
		
		break;
	}

	return 1;
}


int ritas_bc_output(ritas_t *ctx, ritas_bccb_t *bccb, ritas_mbuf_t *m) {

	ritas_bc_abba_header_t *bch;
	u_char value;

	value = *m->head;
	m->head = m->tail - RITAS_BC_ABBA_HEADER_SIZE;
	bch = (ritas_bc_abba_header_t *)m->head;
	if(m->head < m->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		return -1;
	}
	
	bch->type = RITAS_BC_ABBA_PREPROCESS;
	bch->round = 1;
	bch->value = value;

	ritas_bc_abba_share(bccb, bch->share, bch->type, bch->round, bch->value);

	dprintf("[%d] BC out value %d for round %d, type %d\n", bccb->id, bch->value, bch->round, bch->type);

	ritas_rc_broadcast(ctx, bccb->rccb_pre, m);
	
	return 1;	
}

int ritas_bc_abba_verify_share(ritas_bccb_t *bccb, u_char *share, u_short type, u_short round, u_char value, u_short pid) {
	int valid;
	
	ritas_thsig_reset(bccb->thsig, 8);
	ritas_thsig_update(bccb->thsig, (u_char *)&bccb->id, sizeof(bccb->id));
	ritas_thsig_update(bccb->thsig, (u_char *)&type, sizeof(type));
	if(round)
		ritas_thsig_update(bccb->thsig, (u_char *)&round, sizeof(round));
	ritas_thsig_update(bccb->thsig, &value, sizeof(value));
	valid = ritas_thsig_verify_share(bccb->thsig, share, pid);
	return valid;
}
	

int ritas_bc_abba_verify(ritas_bccb_t *bccb, u_char *sig, u_short sig_size, u_short k, u_short type, u_short round, u_char value) {
	int valid;

	ritas_thsig_reset(bccb->thsig, 8);
	ritas_thsig_update(bccb->thsig, (u_char *)&bccb->id, sizeof(bccb->id));
	ritas_thsig_update(bccb->thsig, (u_char *)&type, sizeof(type));			
	if(round)
		ritas_thsig_update(bccb->thsig, (u_char *)&round, sizeof(round));			
	ritas_thsig_update(bccb->thsig, &value, sizeof(value));
	valid = ritas_thsig_verify(bccb->thsig, sig, sig_size, k);
	return valid;
}
	

int ritas_bc_abba_share(ritas_bccb_t *bccb, u_char *buf, u_short type, u_short round, u_char value) {
	int valid;

	ritas_thsig_reset(bccb->thsig, 8);
	ritas_thsig_update(bccb->thsig, (u_char *)&bccb->id, sizeof(bccb->id));
	ritas_thsig_update(bccb->thsig, (u_char *)&type, sizeof(type));			
	ritas_thsig_update(bccb->thsig, (u_char *)&round, sizeof(round));			
	ritas_thsig_update(bccb->thsig, &value, sizeof(value));
	valid = ritas_thsig_generate_share(bccb->thsig, buf);
	return valid;
}

ritas_bccb_t *ritas_bc_init(ritas_t *ctx, u_short id, void *parent) {
	
	ritas_bccb_t *bccb;
	int i;
	
	bccb = malloc(sizeof(*bccb));
	if(!bccb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() control block\n", __func__);
		goto bad;
	}
	memset(bccb, '\0', sizeof(*bccb));
	
	bccb->id = id;
	bccb->round = 1;
	bccb->step = 1;
	
	bccb->decision = RITAS_BC_DEFAULT;

	/*
	 * Check for parent protocol.
	 */
	if(parent) {
		bccb->upper = 1;
		bccb->parent = parent;
	}
	else {
		bccb->upper = 0;
		bccb->parent = NULL;
		//i = ritas_aproxy_init(ctx);
		//if(i == -1)
		//	goto bad;
			
		pthread_mutex_lock(&ctx->pr_mutex);
		ctx->pr_count++;
		pthread_mutex_unlock(&ctx->pr_mutex);			
	}

	/*
	 * Init the prevote justification pointers.
	 */
	bccb->pvjp = &bccb->pvj[0];
	bccb->pvjp_rm1 = &bccb->pvj[1];

	/*
	 * Init Reliable Channel.
	 */
/*
	bccb->rccb = ritas_rc_init(ctx, bccb, RITAS_RC_PROTO_BC);
	if(!bccb->rccb) {
		goto bad;
	}
*/
	bccb->rccb_pre = ritas_rc_init(ctx, bccb, RITAS_RC_PROTO_BC);
	if(!bccb->rccb_pre) {
		goto bad;
	}

	
	/*
	 * Threshold Coin
	 */
	bccb->thcoin = ritas_thcoin_init_bn(ctx->pid, ctx->n, ctx->n - ctx->f);
	if(!bccb->thcoin) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init threshold coin\n", __func__);	 			
		goto bad;
	}
		
	/*
	 * Init Threshold Signatures
	 */
	 bccb->thsig = ritas_thsig_init_ms(ctx->pid, ctx->n);
	 if(!bccb->thsig) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init threshold signatures\n", __func__);	 	
		goto bad;
	 }
	 
	 /*
	  * Init shares
	  */
	for(i = 0 ; i < 2 ; i++) {
	 	bccb->preprocess_share[i] = malloc(ctx->n * sizeof(u_char *));
	 	memset(bccb->preprocess_share[i], '\0', ctx->n * sizeof(u_char *));
	 	bccb->prevote_share[i] = malloc(ctx->n * sizeof(u_char *));
	 	memset(bccb->prevote_share[i], '\0', ctx->n * sizeof(u_char *));
	 	bccb->mainvote_share[i] = malloc(ctx->n * sizeof(u_char *));
		memset(bccb->mainvote_share[i], '\0', ctx->n * sizeof(u_char *));
	}
	bccb->mainvote_share[2] = malloc(ctx->n * sizeof(u_char *));
	memset(bccb->mainvote_share[2], '\0', ctx->n * sizeof(u_char *));

	return bccb;

bad:
	ritas_bc_destroy(ctx, bccb);
	return NULL;

}

void ritas_bc_destroy(ritas_t *ctx, ritas_bccb_t *bccb) {
	int i, j;
	
	if(!bccb)
		return;

	/*
	 * Destroy the prevote justification contents.
	 */
	free(bccb->pvj[0].data);
	free(bccb->pvj[1].data);

	/*
	 * Destroy shares (brrr).
	 */
	for(i = 0 ; i < 3 ; i++) {
		if(i < 2) {
			if(bccb->preprocess_share[i]) {
				for(j = 0 ; j < ctx->n ; j++)
					free(bccb->preprocess_share[i][j]);
				free(bccb->preprocess_share[i]); 	
			}
		
			if(bccb->prevote_share[i]) {
				for(j = 0 ; j < ctx->n ; j++)
					free(bccb->prevote_share[i][j]);
				free(bccb->prevote_share[i]); 	
			}
		}
		
		if(bccb->mainvote_share[i]) {
			for(j = 0 ; j < ctx->n ; j++)
				free(bccb->mainvote_share[i][j]);
			free(bccb->mainvote_share[i]); 	
		}
	}
	
	/*
	 * Destroy threshold signatures.
	 */
	if(bccb->thsig)
		ritas_thsig_destroy(bccb->thsig);
		
	/*
	 * Destroy threshold coin.
	 */
	if(bccb->thcoin)
		ritas_thcoin_destroy(bccb->thcoin);
	
	/*
	 * Destroy prevote justifications.
	 */
	for(i = 0 ; i < 2 ; i++) {
		free(bccb->pvjp->data[i]);
		free(bccb->pvjp_rm1->data[i]);
	}
	
	/*
	 * Destroy reliable channel.
	 */
	if(bccb->rccb)
		ritas_rc_destroy(ctx, bccb->rccb);

	if(bccb->rccb_pre)
		ritas_rc_destroy(ctx, bccb->rccb_pre);

	/*
	 * Destroy aproxy
	 */
	//if(!bccb->upper)
		//ritas_aproxy_destroy(ctx);
		
	/*
	 * Uff, we'good to free() the control block.
	 */
	free(bccb);
}


int ritas_bc(ritas_t *ctx, u_short bcid, u_char proposal) {

	ritas_bccb_t *bccb = NULL;
	ritas_mbuf_t *mbuf = NULL;
	u_char value;
	int n;

	pthread_mutex_lock(&ctx->temp_mutex);
	
	// Init BC control block
	bccb = ritas_bc_init(ctx, bcid, NULL);
	if(!bccb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't initialize bccb\n", __func__);
		goto bad;
	}
	
	// Init mbuf
	mbuf = ritas_mbuf_init();
	if(!mbuf) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't initialize mbuf\n", __func__);
		goto bad;
	}
	
	mbuf->head = mbuf->head - sizeof(proposal);
	if(mbuf->head < mbuf->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		goto bad;
	}
	
	*mbuf->head = proposal;

	// Send the proposal to the BC layer	
	n = ritas_bc_output(ctx, bccb, mbuf);
	if(n == -1)
		goto bad;

	pthread_mutex_unlock(&ctx->temp_mutex);

	// Wait until the BC layer decides
	pthread_mutex_lock(&ctx->aproxy->mutex);
	while(!ctx->aproxy->count)
		pthread_cond_wait(&ctx->aproxy->cond, &ctx->aproxy->mutex);
	pthread_mutex_unlock(&ctx->aproxy->mutex);

	pthread_mutex_lock(&ctx->temp_mutex);

	if(!ctx->aproxy->mvalue)
		goto bad;
	
	value = *ctx->aproxy->mvalue->head;
	//ritas_bc_destroy(ctx, bccb);

	pthread_mutex_unlock(&ctx->temp_mutex);

	return (int)value;
	
bad:
	ritas_bc_destroy(ctx, bccb);
	ritas_mbuf_destroy(mbuf);
	
	return -1;
}


int ritas_bc_multi(ritas_t *ctx, u_short bcid, u_char proposal, u_short n_prop) {

	ritas_bccb_t **bccb = NULL;
	ritas_mbuf_t *mbuf = NULL;
	u_char value;
	int n, i;

	pthread_mutex_lock(&ctx->temp_mutex);
	
	bccb = malloc(n_prop*sizeof(*bccb));
	for(i = 0 ; i < n_prop ; i++) {
		bccb[i] = ritas_bc_init(ctx, bcid+i, NULL);
		if(!bccb) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't initialize bccb\n", __func__);
			goto bad;
		}
	}
	


	gettimeofday(&ctx->t0, NULL);

	// Init BC control blocks	
	for(i = 0 ; i < n_prop ; i++) {
		// Init mbuf
		mbuf = ritas_mbuf_init();
		if(!mbuf) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't initialize mbuf\n", __func__);
			goto bad;
		}		
		
		mbuf->head = mbuf->head - sizeof(proposal);
		if(mbuf->head < mbuf->data) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
			goto bad;
		}

		#ifdef PD_RANDOM
		RAND_pseudo_bytes(mbuf->head, 1);
		*mbuf->head = (*(u_char *)mbuf->head & 0x01 ? 1 : 0);
		#else
		*mbuf->head = proposal;
		#endif
		
		// Send the proposal to the BC layer	
		n = ritas_bc_output(ctx, bccb[i], mbuf);
		if(n == -1)
			goto bad;
			
	}
	pthread_mutex_unlock(&ctx->temp_mutex);

	return 1;
	
bad:
	return -1;
}


u_short ritas_bc_id(ritas_bccb_t *bccb) {
	return bccb->id;
}

void *ritas_bc_parent(ritas_bccb_t *bccb) {
	return bccb->parent;
}
