#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/rand.h>
#include "ritas_bc.h"
#include "ritas_ht.h"
#include "ritas_mbuf.h"
#include "ritas_rb.h"
#include "ritas_eb.h"
#include "ritas_mvc.h"

int ritas_bc_input(ritas_t *ctx, ritas_bccb_t *bccb, ritas_mbuf_t *m) {
	
	ritas_bc_header_t *bch;
	u_short txid;
	int k, n;
	u_char b;

	bch = (ritas_bc_header_t *)m->head;

	bcprintf("%ld us [%d] BC in value %d for round %d, step %d from %d\n", ritas_get_timer(ctx), ctx->pid, bch->value, bch->round, bch->step, m->proc->id);

	if(bccb->terminated) {
		ritas_mbuf_destroy(m);
		dprintf("[%d] BC already terminated\n", bccb->id);
		return 1;
	}
	

	if(bch->phase > bccb->phase) {
		bccb->phase = bch->phase;
		bccb->value = bch->value;
		bccb->status = bch->status;

		memset(bccb->count, sizeof(bccb->count
	};

	
	if(bch->phase == bccb->phase) {
		bccb->count[bch->value]++;
		bccb->count[RITAS_BC_TOTAL]++;
	}

	if



	/*
	 * Update the state.
	 */
	b = bch->value;
	k = bch->step-1;
	
	if(ritas_bc_N(ctx, bccb, k, b)) {
		bccb->valid[k][RITAS_BC_TOTAL]++;
		bccb->valid[k][b]++;

		for(k = k+1 ; k < bccb->step ; k++) {
			if(bccb->invalid[k][b]) {
				if(ritas_bc_N(ctx, bccb, k, b)) {
					bccb->valid[k][RITAS_BC_TOTAL] += bccb->invalid[k][b];
					bccb->valid[k][b] += bccb->invalid[k][b];
					bccb->invalid[k][b] = 0;
				}
			}
		}
		
		// default messages are a special case
		if(bccb->step == 3 && bccb->invalid[RITAS_BC_STEP3][RITAS_BC_DEFAULT]) {
			if(ritas_bc_N(ctx, bccb, RITAS_BC_STEP3, RITAS_BC_DEFAULT)) {
				bccb->valid[RITAS_BC_STEP3][RITAS_BC_TOTAL] += bccb->invalid[RITAS_BC_STEP3][RITAS_BC_DEFAULT];
				bccb->valid[RITAS_BC_STEP3][RITAS_BC_DEFAULT] += bccb->invalid[RITAS_BC_STEP3][RITAS_BC_DEFAULT];
				bccb->invalid[RITAS_BC_STEP3][RITAS_BC_DEFAULT] = 0;
			}
		}	
	}
	else {
		dprintf("[%d] BC invalid!\n", bccb->id);
		bccb->invalid[k][bch->value]++;
		ritas_mbuf_destroy(m);
		return 1;
	}
	
	
	dprintf("[%d] BC {{%d, %d, %d, %d}, {%d, %d, %d, %d}, {%d, %d, %d, %d}}\n", ctx->pid,
	bccb->valid[0][0],bccb->valid[0][1],bccb->valid[0][2],bccb->valid[0][3],
	bccb->valid[1][0],bccb->valid[1][1],bccb->valid[1][2],bccb->valid[1][3],
	bccb->valid[2][0],bccb->valid[2][1],bccb->valid[2][2],bccb->valid[2][3]);

	/*
	 * Check if we have a delayed decision
	 */
	if(!bccb->decided && bccb->round > 1 && bccb->step < 3) {
		if(b != RITAS_BC_DEFAULT && bccb->valid[RITAS_BC_STEP3][b] >= 2*ctx->f+1) {
			dprintf("[%d] BC delayed threshold for round: %d, step %d (2f+1): %d\n", ctx->pid, bch->round, bch->step, b);
			if(!bccb->upper)
				ritas_register_decision_time(ctx->pid, bccb->id, bch->value, bch->round, ctx->t0);
			bccb->decided = 1;
			bccb->terminated = 1;
			
			//ritas_bc_destroy(ctx, bccb);
			
			pthread_mutex_lock(&ctx->pr_mutex);
			ctx->pr_count--;
			pthread_cond_broadcast(&ctx->pr_cond);
			pthread_mutex_unlock(&ctx->pr_mutex);
			
			ritas_mbuf_destroy(m);
			return 1;
		}
	}

	/*
	 * Take the apropriate action according to the new state.
	 */
	if(bccb->valid[bccb->step-1][RITAS_BC_TOTAL] >= ctx->n - ctx->f) {
		switch(bccb->step) {
			/*
			 * Step 1.
			 * Broadcast the majority value from the messages received.
			 */
			case 1:
			bch->value = bccb->valid[RITAS_BC_STEP1][0] > bccb->valid[RITAS_BC_STEP1][1] ? 0 : 1 ;
			
			/*
			 * Lame code block for SRDS paper.
			 */
			if(!bccb->upper) {
				if(bccb->valid[RITAS_BC_STEP1][bch->value] == ctx->n - ctx->f) {
					if(!bccb->decided)
						ritas_register_decision_time(ctx->pid, bccb->id, bch->value, bccb->round, ctx->t0);
					bccb->decided = 1;
				
					//ritas_bc_destroy(ctx, bccb);
					/*
					pthread_mutex_lock(&ctx->pr_mutex);
					ctx->pr_count--;
					pthread_cond_broadcast(&ctx->pr_cond);
					pthread_mutex_unlock(&ctx->pr_mutex);
					*/
				}
			}
						
						
			bcprintf("%ld us [%d] BC threshold for round %d, step %d, MajValue: %d\n", ritas_get_timer(ctx), ctx->pid, bccb->round, bccb->step, bch->value);
			
			bccb->step++;
			bch->step = bccb->step;

			for(txid = 0 ; txid < ctx->n ; txid++) {
				bccb->rbcb[RITAS_BC_STEP2][txid] = ritas_rb_init(ctx, txid, bccb->round*3+bccb->step-3, bccb, RITAS_RB_UPPER_BC, 0);
				if(!bccb->rbcb[RITAS_BC_STEP2][txid])
					return -1;
			}
			
			#ifdef BC_EVIL
			if(ctx->pid < ctx->f)
			bch->value = (bch->value & 0x01 ? 0 : 1);
			#endif			
			
			n = ritas_rb_output(ctx, bccb->rbcb[RITAS_BC_STEP2][ctx->pid], m);
			if(n == -1)
				return -1;
			break;

			/*
			 * Step 2.
			 * If more than half the messages received have the same value, then
			 * broadcast that value. If not, broadcast a zero-valued 
			 * semantically irrelevant message.
			 */
			case 2:
			if(bccb->valid[RITAS_BC_STEP2][0] > ctx->n/2) {
				bch->value = 0;
			}
			else if(bccb->valid[RITAS_BC_STEP2][1] > ctx->n/2) {
				bch->value = 1;
			}
			else {
				bch->value = RITAS_BC_DEFAULT;
			}

			#ifdef BC_EVIL
			if(ctx->pid < ctx->f)
				bch->value = RITAS_BC_DEFAULT;
			#endif

			bcprintf("%ld us [%d] BC threshold for round %d, step %d, >n/2: %d\n", ritas_get_timer(ctx), ctx->pid, bccb->round, bccb->step, bch->value);

			bccb->step++;
			bch->step = bccb->step;
			for(txid = 0 ; txid < ctx->n ; txid++) {
				if(bccb->round > 1) {
					memset(bccb->invalid[RITAS_BC_STEP3], '\0', 3*sizeof(u_short));
					memset(bccb->valid[RITAS_BC_STEP3], '\0', 4*sizeof(u_short));
					n = ritas_rb_renew(ctx, bccb->rbcb[RITAS_BC_STEP3][txid], txid, bccb->round*3+bccb->step-3);
					if(n < 0) {
						return -1;
					}
				}
				else {
					bccb->rbcb[RITAS_BC_STEP3][txid] = ritas_rb_init(ctx, txid, bccb->round*3+bccb->step-3, bccb, RITAS_RB_UPPER_BC, 0);
					if(!bccb->rbcb[RITAS_BC_STEP3][txid])
						return -1;
				}
			}

			n = ritas_rb_output(ctx, bccb->rbcb[RITAS_BC_STEP3][ctx->pid], m);
			if(n == -1)
				return -1;
			break;

			/*
			 * Step 3.
			 * If more than 2f messages have the same value, decide on that 
			 * value.
			 * Otherwise, if more than f messages have the same value, advance
			 * for the next round and broadcast that value.
			 * Otherwise, advance for the next round and broadcast 0 or 1 each
			 * with probability 1/2.
			 */
			case 3:
			bcprintf("%ld us [%d] BC threshold for round: %d, step %d ", ritas_get_timer(ctx), ctx->pid, bccb->round, bccb->step);

			if(bccb->valid[RITAS_BC_STEP3][0] >= 2*ctx->f+1 || bccb->valid[RITAS_BC_STEP3][1] >= 2*ctx->f+1) {
				bch->value = bccb->valid[RITAS_BC_STEP3][0] >= 2*ctx->f+1 ? 0 : 1 ;
				dprintf("(2f+1): %d\n", bch->value);
				
				if(!bccb->decided && !bccb->upper)
					ritas_register_decision_time(ctx->pid, bccb->id, bch->value, bccb->round, ctx->t0);
				bccb->decided = 1;
				bccb->terminated = 1;

				m->head = m->head + RITAS_BC_HEADER_SIZE - sizeof(bch->value);
				if(m->head < m->data) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
					return -1;
				}
				if(!bccb->upper) {
					/*
					 * Deliver to the application
					 */
					/*
					ctx->aproxy->mvalue = m;					 
					pthread_mutex_lock(&ctx->aproxy->mutex);
					ctx->aproxy->count = 1;
					pthread_cond_broadcast(&ctx->aproxy->cond);
					pthread_mutex_unlock(&ctx->aproxy->mutex);
					*/
				
					//ritas_bc_destroy(ctx, bccb);
					pthread_mutex_lock(&ctx->pr_mutex);
					ctx->pr_count--;
					pthread_cond_broadcast(&ctx->pr_cond);					
					pthread_mutex_unlock(&ctx->pr_mutex);
				}
				else {
					/*
					 * Deliver to the parent protocol
					 */
					dprintf("[%d] BC deliver value %d\n", ctx->pid, bch->value);
					m->protocol = RITAS_MBUF_BC;
					n = ritas_mvc_input(ctx, bccb->parent, m);
					if(n == -1)
						return -1;
				}
				return 1;
			}
			else if(bccb->valid[RITAS_BC_STEP3][0] >= ctx->f+1) {
				bch->value = 0;
				#ifdef BC_EVIL
				if(ctx->pid < ctx->f)
					bch->value = 1;
				#endif
	
				dprintf("(f+1): 0\n");
			}
			else if(bccb->valid[RITAS_BC_STEP3][1] >= ctx->f+1) {
				bch->value = 1;
				#ifdef BC_EVIL
				if(ctx->pid < ctx->f)
					bch->value = 0;
				#endif
				dprintf("(f+1): 1\n");
			}
			else {
				RAND_pseudo_bytes(&bch->value, 1);
				bch->value = (bch->value & 0x01 ? 1 : 0);				 
				dprintf("(random): %d\n", bch->value);
			}
			bccb->round++;
			bccb->step = 1;
			for(k = RITAS_BC_STEP1 ; k < RITAS_BC_STEP3 ; k++) {
				memset(bccb->invalid[k], '\0', 3*sizeof(u_short));
				memset(bccb->valid[k], '\0', 4*sizeof(u_short));
			}
			bch->round = bccb->round;
			bch->step = bccb->step;

			/*
			 * TODO: garbage collection?
			 */
			for(txid = 0 ; txid < ctx->n ; txid++) {
				ritas_rb_destroy(ctx, bccb->rbcb[RITAS_BC_STEP2][txid]);
				bccb->rbcb[RITAS_BC_STEP2][txid] = NULL;
				n = ritas_rb_renew(ctx, bccb->rbcb[RITAS_BC_STEP1][txid], txid, bccb->round*3+bccb->step-3);
				if(n == -1)
					return -1;
			}

			n = ritas_rb_output(ctx, bccb->rbcb[RITAS_BC_STEP1][ctx->pid], m);
			if(n == -1)
				return -1;
		}		
	}
	else {
		ritas_mbuf_destroy(m);
	}

	return 1;			
}


int ritas_bc_output(ritas_t *ctx, ritas_bccb_t *bccb, ritas_mbuf_t *m) {
	
	ritas_bc_header_t *bch;

	/*
	 * Build BC Header
	 */
	m->head = m->head - RITAS_BC_HEADER_SIZE + sizeof(u_char);
	if(m->head < m->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		return -1;
	}
	bch = (ritas_bc_header_t *)m->head;
	bch->step = 1;
	bch->round = 1;
	
	dprintf("[%d] BC out value %d for round %d, step %d\n", ctx->pid, bch->value, bch->round, bch->step);
	
	/*
	 * Reliably Broadcast the message.
	 */
	return ritas_rb_output(ctx, bccb->rbcb[RITAS_BC_STEP1][ctx->pid], m);
}	


ritas_bccb_t *ritas_bc_init(ritas_t *ctx, u_short id, void *parent) {
	
	ritas_bccb_t *bccb;
	u_short txid;
	int i;
	
	bccb = malloc(sizeof(ritas_bccb_t));
	if(!bccb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't malloc() bccb\n", __func__);
		goto bad;
	}
	memset(bccb, '\0', sizeof(*bccb));

	/*
	 * Check for parent protocol.
	 */
	if(parent) {
		bccb->upper = 1;
		bccb->parent = parent;
	}
	else {
		bccb->upper = 0;
		bccb->parent = NULL;
		/*
		i = ritas_aproxy_init(ctx);
		if(i == -1)
			goto bad;
		*/
		pthread_mutex_lock(&ctx->pr_mutex);
		ctx->pr_count++;
		pthread_mutex_unlock(&ctx->pr_mutex);			
	}
		
	
	bccb->id = id;
	bccb->round = 1;
	bccb->step = 1;

	for(i = 0 ; i < 3 ; i++) {
		bccb->rbcb[i] = malloc(ctx->n*sizeof(*bccb->rbcb[i]));
		if(!bccb->rbcb[i]) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't malloc() bccb->rbcb\n", __func__);
			goto bad;
		}
		memset(bccb->rbcb[i], '\0', ctx->n*sizeof(*bccb->rbcb[i]));
	
		if(i == 0) {
			for(txid = 0 ; txid < ctx->n ; txid++) {
				bccb->rbcb[i][txid] = ritas_rb_init(ctx, txid, bccb->round*3+bccb->step-3, bccb, RITAS_RB_UPPER_BC, 0);
				if(!bccb->rbcb[i][txid])
					goto bad;
			}
		}
	}

	return bccb;

bad:
	ritas_bc_destroy(ctx, bccb);
	return NULL;
}

void ritas_bc_destroy(ritas_t *ctx, ritas_bccb_t *bccb) {
	
	u_short txid;
	int i;
	
	if(bccb) {
		if(bccb->rbcb) {
			for(i = 0 ; i < 3 ; i++) {
				for(txid = 0 ; txid < ctx->n ; txid++)
					ritas_rb_destroy(ctx, bccb->rbcb[i][txid]);
				free(bccb->rbcb[i]);
			}	
		}
		
		if(!bccb->upper) {
			ritas_aproxy_destroy(ctx);
		}
		free(bccb);
	}
}

int ritas_bc(ritas_t *ctx, u_short bcid, u_char proposal) {

	ritas_bccb_t *bccb = NULL;
	ritas_mbuf_t *mbuf = NULL;
	u_char value;
	int n;

	pthread_mutex_lock(&ctx->temp_mutex);
	
	// Init BC control block
	bccb = ritas_bc_init(ctx, bcid, NULL);
	if(!bccb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't initialize bccb\n", __func__);
		goto bad;
	}
	
	// Init mbuf
	mbuf = ritas_mbuf_init();
	if(!mbuf) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't initialize mbuf\n", __func__);
		goto bad;
	}
	
	mbuf->head = mbuf->head - sizeof(proposal);
	if(mbuf->head < mbuf->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		goto bad;
	}
	
	*mbuf->head = proposal;

	// Send the proposal to the BC layer	
	n = ritas_bc_output(ctx, bccb, mbuf);
	if(n == -1)
		goto bad;

	pthread_mutex_unlock(&ctx->temp_mutex);

	// Wait until the BC layer decides
	pthread_mutex_lock(&ctx->aproxy->mutex);
	while(!ctx->aproxy->count)
		pthread_cond_wait(&ctx->aproxy->cond, &ctx->aproxy->mutex);
	pthread_mutex_unlock(&ctx->aproxy->mutex);

	pthread_mutex_lock(&ctx->temp_mutex);

	if(!ctx->aproxy->mvalue)
		goto bad;
	
	value = *ctx->aproxy->mvalue->head;
	ritas_bc_destroy(ctx, bccb);

	pthread_mutex_unlock(&ctx->temp_mutex);

	return (int)value;
	
bad:
	ritas_bc_destroy(ctx, bccb);
	ritas_mbuf_destroy(mbuf);
	
	return -1;
}

u_short ritas_bc_id(ritas_bccb_t *bccb) {
	return bccb->id;
}

void *ritas_bc_parent(ritas_bccb_t *bccb) {
	return bccb->parent;
}

int ritas_bc_N(ritas_t *ctx, ritas_bccb_t *bccb, u_short k, u_char b) {
	switch(k) {
		case RITAS_BC_STEP1:
		return 1;
		
		case RITAS_BC_STEP2:
		if(bccb->valid[RITAS_BC_STEP1][b] > (ctx->n - ctx->f)/2)
			return 1;
		else
			return 0;
			
		case RITAS_BC_STEP3:
		if(bccb->round > 1 && bccb->step < 3)
			return 1;
		
		if(b == RITAS_BC_DEFAULT)
			if(bccb->valid[RITAS_BC_STEP2][0] >= (ctx->n - ctx->f) - ctx->n/2
			&& bccb->valid[RITAS_BC_STEP2][1] >= (ctx->n - ctx->f) - ctx->n/2)
				return 1;
			else
				return 0;
		else if(bccb->valid[RITAS_BC_STEP2][b] > (ctx->n - ctx->f)/2)
			return 1;
		else
			return 0;
	}
}


int ritas_bc_multi(ritas_t *ctx, u_short bcid, u_char proposal, u_short n_prop) {

	ritas_bccb_t **bccb = NULL;
	ritas_mbuf_t *mbuf = NULL;
	u_char value;
	int n, i;

	pthread_mutex_lock(&ctx->temp_mutex);
	
	bccb = malloc(n_prop*sizeof(*bccb));
	for(i = 0 ; i < n_prop ; i++) {
		bccb[i] = ritas_bc_init(ctx, bcid+i, NULL);
		if(!bccb) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't initialize bccb\n", __func__);
			goto bad;
		}
	}
		
	gettimeofday(&ctx->t0, NULL);
	ritas_reset_timer(ctx);

	// Init BC control blocks	
	for(i = 0 ; i < n_prop ; i++) {
		// Init mbuf
		mbuf = ritas_mbuf_init();
		if(!mbuf) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't initialize mbuf\n", __func__);
			goto bad;
		}

		mbuf->head = mbuf->head - sizeof(proposal);
		if(mbuf->head < mbuf->data) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
			goto bad;
		}

		#ifdef PD_RANDOM
		RAND_pseudo_bytes(mbuf->head, 1);
		*mbuf->head = (*(u_char *)mbuf->head & 0x01 ? 1 : 0);
		#else
		*mbuf->head = proposal;
		#endif
		
		n = ritas_bc_output(ctx, bccb[i], mbuf);
		if(n == -1)
			goto bad;
	}
	pthread_mutex_unlock(&ctx->temp_mutex);

	return 1;

bad:
	return -1;
}
