#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ritas.h"
#include "ritas_eb.h"
#include "ritas_rc.h"
#include "ritas_bc.h"
#include "ritas_mvc.h"
#include "ritas_vc.h"
#include "ritas_ab.h"
#include "ritas_netio.h"

int ritas_eb_input(ritas_t *ctx, ritas_ebcb_t *ebcb, ritas_mbuf_t *m) {
	
	ritas_eb_header_t *ebh;
	u_char *ebh_value;
	u_short ebh_value_s;
	ritas_vector_t *vector;
	int i, j, n;
	u_char valid;
	MD5_CTX md5_ctx;
	u_char hash[MD5_DIGEST_LENGTH];
	

	ebh = (ritas_eb_header_t *)m->head;
	ebh_value = m->head + RITAS_EB_HEADER_SIZE;
	ebh_value_s = m->tail - ebh_value;
	
	dprintf("[%d] EB in type %d from %d {%d,%d} step %d, count %d\n", ctx->pid, ebh->type, m->proc->id, ebcb->txid, ebcb->rbid, ebcb->step, ebcb->count);	
	/*
	 * Check if the message is still relevant for our step.
	 */
	if(ebcb->step == 2) {
		if(ebcb->role != RITAS_EB_SENDER)
			if(ebh->type != RITAS_EB_COMMIT) {
				dprintf("[%d] EB in BAD 1 from %d\n", ctx->pid, m->proc->id);
				ritas_mbuf_destroy(m);
				return 1;
			}
		else {
			if(ebh->type == RITAS_EB_INITIAL) {
				dprintf("[%d] EB in BAD 2 from %d\n", ctx->pid, m->proc->id);
				ritas_mbuf_destroy(m);
				return 1;
			}
		}
	}
	else if(ebcb->step > 2) {
		dprintf("[%d] EB in BAD 3 from %d\n", ctx->pid, m->proc->id);		
		ritas_mbuf_destroy(m);
		return 1;
	}
			

	switch(ebh->type) {
		case RITAS_EB_INITIAL:
			dprintf("[%d] EB in INITIAL from %d\n", ctx->pid, m->proc->id);
			
			/*
			 * It's a INITIAL message. Build the hash vector and send it to the
			 * sender.
			 */
			ebcb->m = m;
			m = ritas_mbuf_init();
			
			m->head = m->head - ctx->n*MD5_DIGEST_LENGTH;
			for(i = 0 ; i < ctx->n ; i++) {
				MD5_Init(&md5_ctx);
				MD5_Update(&md5_ctx, ebh_value, ebh_value_s);
				MD5_Update(&md5_ctx, ctx->group[i]->key, ctx->group[i]->key_s);
				MD5_Final(m->head + i*MD5_DIGEST_LENGTH, &md5_ctx);
			}
			
			m->head = m->head - RITAS_EB_HEADER_SIZE;
			ebh = (ritas_eb_header_t *)m->head;
			ebh->type = RITAS_EB_ECHO;
			
			n = ritas_rc_send(ctx, ebcb->rccb, m, ctx->group[ebcb->txid]);
			if(n == -1)
				return -1;
			
			ebcb->step++;
			break;					

		case RITAS_EB_ECHO:
			if(ebcb->role != RITAS_EB_SENDER || ebcb->count >= ctx->n - ctx->f) {
				ritas_mbuf_destroy(m);
				return 1;
			}

			dprintf("[%d] EB in ECHO from %d\n", ctx->pid, m->proc->id);
			/*
			 * Check if we already have the hash vector from this process.
			 */
			if(ebcb->matrix[m->proc->id]) {
				ritas_mbuf_destroy(m);
				return 1;
			}
		
			/*
			 * Add vector to the hash matrix and increment count.
			 */
			ebcb->matrix[m->proc->id] = malloc(ebh_value_s);
			if(!ebcb->matrix[m->proc->id]) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): malloc()\n", __func__);
				return -1;
			}
			memcpy(ebcb->matrix[m->proc->id], ebh_value, ebh_value_s);
			
			ebcb->count++;
		
			/*
			 * Check if we reached the threshold.
			 */
			if(ebcb->count == ctx->n - ctx->f) {
				dprintf("[%d] EB threshold!\n", ctx->pid);
				/*
				 * Send to each process their corresponding column in the matrix.
				 */
				ebh->type = RITAS_EB_COMMIT;
				m->flags |= RITAS_MBUF_PROTECTED;
				for(j = 0 ; j < ctx->n ; j++) {
					vector = ritas_vector_init(ctx->n);
					if(!vector) {
						snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init vector\n", __func__);
						return -1;
					}
					
					for(i = 0 ; i < ctx->n ; i++) {
						if(ebcb->matrix[i]) {
							n = ritas_vector_add(vector, ebcb->matrix[i] + j*MD5_DIGEST_LENGTH, MD5_DIGEST_LENGTH);
						}
						else
							n = ritas_vector_add(vector, NULL, 0);
						if(n == -1) {
							snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add element to vector\n", __func__);
							return -1;
						}
					}
					ritas_vector_build(vector, ebh_value); // TODO: bounds checking
					m->head = (u_char *)ebh;
					m->tail = ebh_value + vector->data_s;
					n = ritas_rc_send(ctx, ebcb->rccb, m, ritas_proc_get(ctx, j));
					if(n == -1)
						return -1;
						
					ritas_vector_destroy(vector);
				}
				m->flags ^= RITAS_MBUF_PROTECTED;
				ritas_mbuf_destroy(m);
			}				
			break;
	
		case RITAS_EB_COMMIT:
			dprintf("[%d] EB in COMMIT from %d\n", ctx->pid, m->proc->id);

			/*
			 * It's an COMMIT message. Validate the vector and deliver the
			 * message.
			 */
			vector = ritas_vector_extract(ebh_value, ebh_value_s);
			if(!vector) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to extract vector\n", __func__);
				return -1;
			}

			valid = 0;
			for(i = 0 ; i < ctx->n ; i++) {
 				MD5_Init(&md5_ctx);
			 	MD5_Update(&md5_ctx, ebcb->m->head + RITAS_EB_HEADER_SIZE, ebcb->m->tail - ebcb->m->head - RITAS_EB_HEADER_SIZE);
			 	MD5_Update(&md5_ctx, ctx->group[i]->key, ctx->group[i]->key_s);
			 	MD5_Final(hash, &md5_ctx);
			 	if(!memcmp(hash, vector->element[i], vector->element_s[i])) {
			 		valid++;
			 	}
			}

			if(valid >= ctx->n - ctx->f) {
				ebcb->step++;
				ebcb->m->protocol = RITAS_MBUF_RB;
				ebcb->m->head += RITAS_EB_HEADER_SIZE;
				if(!ebcb->upper) {
					/*
					 * Deliver to the application
					 */
					ctx->aproxy->mvalue = ebcb->m;
					pthread_mutex_lock(&ctx->aproxy->mutex);
					ctx->aproxy->count++;
					pthread_cond_broadcast(&ctx->aproxy->cond);
					pthread_mutex_unlock(&ctx->aproxy->mutex);
				}
				else {
					/*
					 * Pass to the upper layer
					 */
					switch(ebcb->upper) {
					case RITAS_RB_UPPER_BC:
						i = ritas_bc_input(ctx, (ritas_bccb_t *)ebcb->parent, ebcb->m);
						break;
					case RITAS_RB_UPPER_MVC:
						i = ritas_mvc_input(ctx, (ritas_mvccb_t *)ebcb->parent, ebcb->m);
						break;
					case RITAS_RB_UPPER_VC:
						i = ritas_vc_input(ctx, (ritas_vccb_t *)ebcb->parent, ebcb->m);
						break;
					case RITAS_RB_UPPER_AB:
						i = ritas_ab_input(ctx, (ritas_abcb_t *)ebcb->parent, ebcb->m);
						break;
					}
				}
				if (i == -1)
					return -1;
			}
			break;		
	}

	return 1;
}	

int ritas_eb_output(ritas_t *ctx, ritas_ebcb_t *ebcb, ritas_mbuf_t *m) {
	
	ritas_eb_header_t *ebh;
	
	/*
	 * Build EB Header
	 */
	m->head = m->head - RITAS_EB_HEADER_SIZE;
	if(m->head < m->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		return -1;
	}
	ebh = (ritas_eb_header_t *)m->head;
	ebh->type = RITAS_EB_INITIAL;
	
	/*
	 * Broadcast the message over the reliable channel
	 */
	return ritas_rc_broadcast(ctx, ebcb->rccb, m);
}

ritas_ebcb_t *ritas_eb_init(ritas_t *ctx, u_short txid, u_short id, void *parent, u_char parent_type) {

	ritas_ebcb_t *ebcb;
	int i;
	
	/*
	 * Create EBCB
	 */
	ebcb = (ritas_ebcb_t *)malloc(sizeof(*ebcb));
	if(!ebcb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() control block\n", __func__);
		goto bad;
	}
	memset(ebcb, '\0', sizeof(*ebcb));

	/*
	 * Check for parent protocol.
	 */
	if(parent) {
		if(parent_type < 1 || parent_type > 4) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): invalid parent type\n", __func__);
			goto bad;
		}
		ebcb->upper = parent_type;
		ebcb->parent = parent;
	}
	else {
		ebcb->upper = 0;
		i = ritas_aproxy_init(ctx);
		if(i == -1)
			goto bad;
	}
	
	ebcb->txid = txid;
	ebcb->rbid = id;
	ebcb->step = 1;
	
	/*
	 * Define role.
	 */
	if(txid == ctx->pid)
		ebcb->role = RITAS_EB_SENDER;
	else
		ebcb->role = RITAS_EB_RECEIVER;

	/*
	 * Init matrix.
	 */
	if(ebcb->role == RITAS_EB_SENDER) {
		ebcb->matrix = malloc(ctx->n*sizeof(u_char *));
		if(!ebcb->matrix) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() matrix\n", __func__);
			goto bad;
		}
		memset(ebcb->matrix, '\0', ctx->n*sizeof(u_char *));
	}
	
	ebcb->rccb = ritas_rc_init(ctx, ebcb, RITAS_RC_PROTO_EB);
	if(!ebcb->rccb) {
		goto bad;
	}
	
	return ebcb;
	
bad:
	if(ebcb)
		ritas_eb_destroy(ctx, ebcb);
	
	return NULL;
}

void ritas_eb_destroy(ritas_t *ctx, ritas_ebcb_t *ebcb) {

	int i;

	if(!ebcb->upper)
		ritas_aproxy_destroy(ctx);

	if(ebcb->role == RITAS_EB_SENDER) {
		for(i = 0 ; i < ctx->n ; i++) {
			free(ebcb->matrix[i]);
		}
		free(ebcb->matrix);
	}

	//ritas_mbuf_destroy(ebcb->m);
	ritas_rc_destroy(ctx, ebcb->rccb);

	free(ebcb);
}

ritas_eb_renew(ritas_t *ctx, ritas_ebcb_t *ebcb, u_short txid, u_short id) {
	
	int i;
	
	ebcb->txid = txid;
	ebcb->rbid = id;
	ebcb->step = 1;
	ebcb->count = 0;

	dprintf("[%d] EB renew {%d,%d} step %d, count %d\n", ctx->pid, ebcb->txid, ebcb->rbid, ebcb->step, ebcb->count);
	
	/*
	 * Define role.
	 */
	if(txid == ctx->pid)
		ebcb->role = RITAS_EB_SENDER;
	else
		ebcb->role = RITAS_EB_RECEIVER;	

	//ritas_mbuf_destroy(ebcb->m);
	ebcb->m = NULL;
	
	/*
	 * Renew matrix.
	 */
	if(ebcb->role == RITAS_EB_SENDER) {
		for(i = 0 ; i < ctx->n ; i++) {
			free(ebcb->matrix[i]);
			ebcb->matrix[i] = NULL;
		}
	}	
	
	i = ritas_rc_renew(ctx, ebcb->rccb, ebcb);
	if(i == -1) {
		goto bad;
	}
	
	return 1;
	
bad:
	ritas_eb_destroy(ctx, ebcb);
	return -1;
}

u_short ritas_eb_txid(ritas_ebcb_t *ebcb) {
	return ebcb->txid;
}

u_short ritas_eb_rbid(ritas_ebcb_t *ebcb) {
	return ebcb->rbid;
}

void *ritas_eb_parent(ritas_ebcb_t *ebcb) {
	return ebcb->parent;
}

u_char ritas_eb_parent_type(ritas_ebcb_t *ebcb) {
	return ebcb->upper;
}
