#ifndef RITAS_HT_H
#define RITAS_HT_H

#include <sys/types.h>
#ifdef MEMWATCH
#include "memwatch.h"
#endif

/*
 * Generic hash table.
 * To be used as basis for more specific hash tables.
 */
struct ritas_ht {
	u_char size;
	struct ritas_ht_entry **table;
};
typedef struct ritas_ht ritas_ht_t;

/*
 * Hash table entry.
 */
struct ritas_ht_entry {
	u_char *key;
	u_short ksize;
	void *value;
	struct ritas_ht_entry *next;
};
typedef struct ritas_ht_entry ritas_ht_entry_t;


ritas_ht_t *ritas_ht_init(u_char tsize);
void ritas_ht_destroy(ritas_ht_t *ht, void (*destroy_func)(void *v));
void ritas_ht_renew(ritas_ht_t *ht, void (*destroy_func)(void *v));
int ritas_ht_add(ritas_ht_t *ht, u_char *key, u_short ksize, void *value);
int ritas_ht_add_ext(ritas_ht_t *ht, u_char *key, u_short ksize, void *value);
void *ritas_ht_remove(ritas_ht_t *ht, u_char *key, u_short ksize);
ritas_ht_entry_t *ritas_ht_remove_ext(ritas_ht_t *ht, u_char *key, u_short ksize);
void *ritas_ht_lookup(ritas_ht_t *ht, u_char *key, u_short ksize);
ritas_ht_entry_t *ritas_ht_lookup_ext(ritas_ht_t *ht, u_char *key, u_short ksize);
u_char ritas_ht_hash(ritas_ht_t *ht, u_char *key, u_short ksize);
#endif	/* RITAS_HT_T */
