#include <stdlib.h>
#include "ritas_list.h"
#include "ritas_mbuf.h"

ritas_list_t *ritas_list_init() {
	
	ritas_list_t *l;
	
	l = malloc(sizeof(*l));
	if(!l)
	    return NULL;
	
	l->first = l->last = NULL;
	
	return l;
}
	
void ritas_list_destroy(ritas_list_t *l) {
	if(l)
		free(l);
}

void ritas_list_destroy_mbuf(ritas_list_t *l) {
	ritas_mbuf_t *m;
	
	while((m = ritas_list_dequeue(l)))
		ritas_mbuf_destroy(m);
	free(l);
}

int ritas_list_add(ritas_list_t *l, void *t, void *o) {
	
	ritas_list_member_t *m;
	ritas_list_member_t *target;
	
	if(!l)
		 return -1;
	
	if(!o) {
		return -1;
	}

	target = (ritas_list_member_t *)t;
	m = (ritas_list_member_t *)o;

	if(!target) {
		m->prev = NULL;
		m->next = l->first;
		if(l->first)
			l->first->prev = m;
		else
			l->last = m;
		l->first = m;
		
	}
	else {
		m->prev = target;
		m->next = target->next;
		target->next = m;
		if(l->last == target)
			l->last = m;
	}

	return 1;
}


void *ritas_list_remove(ritas_list_t *l, void *target) {
	
	ritas_list_member_t *m;
	
	if(!l)
		return NULL;
	
	if(!target)
		return NULL;
	
	m = (ritas_list_member_t *)target;

	if(!m->prev)
		l->first = m->next;
	else
		m->prev->next = m->next;

	if(!m->next)
		l->last = m->prev;
	else
		m->next->prev = m->prev;
	
	m->prev = m->next = NULL;
	
	return (void *)m;
}

int ritas_list_enqueue(ritas_list_t *l, void *o) {
	
	if(!l)
		return 1;
	
	if(!o)
		return 1;
	
	return ritas_list_add(l, l->last, o);
}

void *ritas_list_dequeue(ritas_list_t *l) {

	if(!l)
		return NULL;
	
	if(!l->first)
		return NULL;
	
	return ritas_list_remove(l, l->first);
}

void *ritas_list_concat(ritas_list_t *l1, ritas_list_t *l2) {
	

	if(!l1->last) {
		l1->first = l2->first;
		l1->last = l2->last;
	} 
	else if(!l2->last) {
		// do nothing
	}
	else {
		l1->last->next = l2->first;
	}

	ritas_list_destroy(l2);
	
	return l1;
}
