#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ritas_mvc.h"
#include "ritas_vc.h"
#include "ritas_bc.h"
#ifdef ECHO
#include "ritas_eb.h"
#endif
#include "ritas_rb.h"
#include "ritas_vector.h"

int ritas_mvc_input(ritas_t *ctx, ritas_mvccb_t *cb, ritas_mbuf_t *m) {

	ritas_mvc_header_t *mvch;
	ritas_mvc_val_t *init_count;
	u_short *vect_count, *vindex;
	u_char *v;
	u_short v_size;
	int i, n;
	ritas_vector_t *vector;
	ritas_ht_entry_t *e;
	ritas_mbuf_t *m_aux;

	/*
	 * Check message protocol
	 */
	if(m->protocol == RITAS_MBUF_BC) {
		/*
		 * Process consensus result
		 */
		cb->bc_decision = (u_char)*m->head;
		ritas_mbuf_destroy(m);
		mvcprintf("%ld us [%d] MVC(%d) decision BC value %d\n", ritas_get_timer(ctx), ctx->pid, cb->id, cb->bc_decision);
	}
	else {
		mvch = (ritas_mvc_header_t *)m->head;
		v_size = m->tail - m->head - RITAS_MVC_HEADER_SIZE;

		/*
		 * Process RB message.
		 */
	 	if(mvch->type == RITAS_MVC_INIT) {
	 		v = m->head + RITAS_MVC_HEADER_SIZE;
	 	
	 		mvcprintf("%ld us [%d] MVC(%d) in INIT from %d, u_int value %d (%d bytes)\n", ritas_get_timer(ctx), ctx->pid, cb->id, m->proc->id, *(u_int *)v, v_size);
	 	
	 		/*
	 		 * Update INIT value table.
	 		 */
	 		init_count = ritas_ht_lookup(cb->vtable[RITAS_MVC_INIT], v, v_size);
	 		if(!init_count) {
	 			/*
	 			 * It's a new value.
	 			 */
	 			init_count = malloc(sizeof(ritas_mvc_val_t));
	 			if(!init_count) {
	 				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() init count\n", __func__);
	 				return -1;
	 			}
	 			init_count->count = 0;
	 			init_count->vect_count = 0;
	 			
	 			i = ritas_ht_add(cb->vtable[RITAS_MVC_INIT], v, v_size, init_count);
	 			if(i == -1) {
	 				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to hash table\n", __func__);
	 				return -1;
	 			}
	 		}
	 		e = ritas_ht_lookup_ext(cb->vtable[RITAS_MVC_INIT], v, v_size);
 			if(!e) {
 				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable locate entry in hash table\n", __func__);
 				return -1;
 			}
 			cb->init_vproc_key[m->proc->id] = e->key;
 			cb->init_vproc_key_size[m->proc->id] = v_size; 
 			cb->init_vproc_key_total_size += v_size; 		
	 		
			init_count->count++;	 			
			if(init_count->count > cb->hcount[RITAS_MVC_INIT]) {
				if(cb->hcount_key[RITAS_MVC_INIT] != e->key) {
					cb->hcount_key[RITAS_MVC_INIT] = e->key;
					cb->hcount_key_size[RITAS_MVC_INIT] = v_size;
				}
				cb->hcount[RITAS_MVC_INIT] = init_count->count;
			}
			cb->tcount[RITAS_MVC_INIT]++;


			/*
			 * Have we reached the n-f threshold?
			 */
			if(cb->tcount[RITAS_MVC_INIT] == ctx->n - ctx->f) {
				mvcprintf("%ld us [%d] MVC(%d) threshold INIT\n", ritas_get_timer(ctx), ctx->pid, cb->id);
				/*
				 * We are ready to start receiving VECT messages.
				 * Initialize the corresponding control blocks.
				 */
					for(i = 0 ; i < ctx->n ; i++) {
#ifdef ECHO
						cb->vect_cb[i] = ritas_eb_init(ctx, i, 2, cb, RITAS_MVC);
#else
						cb->vect_cb[i] = ritas_rb_init(ctx, i, 2, cb, RITAS_MVC, RITAS_RB_FLG_ECHO);
#endif
						if(!cb->vect_cb[i])
							return -1;
					}
					
					/*
				 * Build VECT message.
				 */
				m->head = m->tail = m->data + RITAS_MBUF_DATA_SIZE; // reset mbuf
				if(m->head < m->data) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
					return -1;
				}
				
				#ifdef EVIL
				if(ctx->pid == 0)
					cb->hcount[RITAS_MVC_INIT] = 1;
				#endif
				if(cb->hcount[RITAS_MVC_INIT] >= ctx->n - 2*ctx->f) {
					vector = ritas_vector_init(ctx->n);
					if(!vector) {
						snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init vector\n", __func__);
						return -1;
					}
				
					vindex = malloc(ctx->n*sizeof(u_short));

					for(i = 0 ; i < ctx->n ; i++) {
						if(cb->init_vproc_key[i] == cb->hcount_key[RITAS_MVC_INIT]) {
							vindex[i] = i;
							n = ritas_vector_add(vector, (u_char *)&vindex[i], sizeof(*vindex));
							if(n == -1) {
								snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to vector\n", __func__);
								return -1;
							}
						}
					}
					
					#ifdef DEBUG
					dprintf("[%d] MVC built vector (%d bytes): ", ctx->pid, vector->data_s);
					for(i = 0 ; i < vector->size ; i++)
						dprintf("(%d) ", *(u_short *)vector->element[i]);
					dprintf("\n");
					#endif

					m->head = m->head - vector->data_s;
					if(m->head < m->data) {
						snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
						return -1;
					}				
					ritas_vector_build(vector, m->head);
					ritas_vector_destroy(vector);
					free(vindex);
				}
				
				/*
				 * Build the header.
				 */
				m->head = m->head - RITAS_MVC_HEADER_SIZE;
				if(m->head < m->data) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
					return -1;
				}
				mvch = (ritas_mvc_header_t *)m->head;
				mvch->type = RITAS_MVC_VECT;

				/*
				 * Yey!
				 */
				cb->step = 2;
#ifdef ECHO
				n = ritas_eb_output(ctx, cb->vect_cb[ctx->pid], m);
#else
				n = ritas_rb_output(ctx, cb->vect_cb[ctx->pid], m);
#endif				
				if(n == -1)
					return -1;
	 		}
			else
				ritas_mbuf_destroy(m);
	 		
			if(cb->step == 2) {
				cb->incomplete = 1;
			}
	 	}
	 	else if(mvch->type == RITAS_MVC_VECT) {
	 		/*
	 		 * Add it to incomplete.
	 		 */
	 		mvcprintf("%ld us [%d] MVC(%d) in VECT from %d, size %d bytes\n", ritas_get_timer(ctx), ctx->pid, cb->id, m->proc->id, m->tail - m->head);
			ritas_list_enqueue(cb->incomplete_l, m);
	 		cb->incomplete = 1;
	 	}
	 	else
	 		ritas_mbuf_destroy(m);

	 	
	 	/*
	 	 * Validate messages in incomplete list.
	 	 */
		if(cb->incomplete && cb->step == 2) {
			m_aux = (ritas_mbuf_t *)cb->incomplete_l->first;
			while(m_aux) {
				n = ritas_mvc_validate(ctx, cb, m_aux, &v, &v_size);
				if(n == -1) {
					dprintf("ARGH\n");
					ritas_mbuf_destroy(m_aux);
					return -1;
				}
				
				if(n == RITAS_MVC_INVALID) {
			 		dprintf("WOWOHA, A MESSAGE IS INVALID!\n");
			 		m = m_aux;
					m_aux = m_aux->next;
			 		ritas_list_remove(cb->incomplete_l, m);
			 		ritas_mbuf_destroy(m);
				}
				else if(n == RITAS_MVC_INCOMPLETE) {
			 		dprintf("WOWOHA, A MESSAGE IS INCOMPLETE!\n");
			 		m_aux = m_aux->next;
				}
				else { // n == RITAS_MVC_VALID
		 			m = m_aux;
		 			m_aux = m_aux->next;
		 			ritas_list_remove(cb->incomplete_l, m);
		 			if(v_size > 0) {
				 		/*
				 		 * Update VECT value table.
				 		 */
				 		vect_count = ritas_ht_lookup(cb->vtable[RITAS_MVC_VECT], v, v_size);
				 		if(!vect_count) {
				 			/*
				 			 * It's a new value.
				 			 */
				 			vect_count = malloc(sizeof(u_short));
				 			*vect_count = 0;
	
				 			n = ritas_ht_add(cb->vtable[RITAS_MVC_VECT], v, v_size, vect_count);
			 				if(n == -1) {
				 				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to hash table\n", __func__);
			 					return -1;
				 			}
			 			}
				 		e = ritas_ht_lookup_ext(cb->vtable[RITAS_MVC_INIT], v, v_size);
	 					if(!e) {
		 					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable locate entry in hash table\n", __func__);
	 						return -1;
			 			}
			 			
						(*vect_count)++;	 			
						if(*vect_count > cb->hcount[RITAS_MVC_VECT]) {
							if(cb->hcount_key[RITAS_MVC_VECT] != e->key) {
								cb->hcount_key[RITAS_MVC_VECT] = e->key;
								cb->hcount_key_size[RITAS_MVC_VECT] = v_size;
							}
							cb->hcount[RITAS_MVC_VECT] = *vect_count;
						}
			 		}
		 			else {
		 				/*
			 			 * It's a default value!
			 			 */
			 			cb->vect_dcount++;
			 		}
					cb->tcount[RITAS_MVC_VECT]++;
	
					if(cb->step <= 2) {
						/*
						 * Have we reached the n-f threshold?
						 */
						if(cb->tcount[RITAS_MVC_VECT] >= ctx->n - ctx->f) {
							mvcprintf("%ld us [%d] MVC(%d) threshold VECT\n", ritas_get_timer(ctx), ctx->pid, cb->id);
							m->tail = m->data + RITAS_MBUF_DATA_SIZE;
							m->head = m->tail - sizeof(u_char);
							if(m->head < m->data) {
								snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
								return -1;
							}
							if(cb->tcount[RITAS_MVC_VECT] == cb->vect_dcount + cb->hcount[RITAS_MVC_VECT] && cb->hcount[RITAS_MVC_VECT] >= ctx->n - 2*ctx->f)
								*m->head = 1;
							else 
								*m->head = 0;
				
							#ifdef EVIL
							if(ctx->pid == 0)
								*m->head = 0;
							#endif
							
							cb->step = 3;
							n = ritas_bc_output(ctx, cb->bccb, m);
							if(i == -1)
				 				return -1;
				 			break;
			 			}
			 			else
			 				ritas_mbuf_destroy(m);
					}
					else
		 				ritas_mbuf_destroy(m);
		 		}
			}
		}
		cb->incomplete = 0;

	}
	
 	/*
 	 * It's time to deliver.
 	 * We're in the last step and have a BC decision which is: zero; or one and
 	 * the number of valid VECT messages with the same value is >= n-2f.
 	 */
 	if(cb->step == 3 && cb->bc_decision != RITAS_MVC_BC_UNDECIDED
 	&&(cb->bc_decision == 0 || cb->hcount[RITAS_MVC_VECT] >= ctx->n - 2*ctx->f)) {
		/*
		 * Build mbuf for upper protocol. A zero-length mbuf stands for the 
		 * default value.
		 */
		m = ritas_mbuf_init();
		if(!m) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init mbuf\n", __func__);
			return -1;
		}

		if(cb->bc_decision == 1) {
			/*
			 * Copy the decided value to the mbuf.
			 */

			m->tail = m->data + RITAS_MBUF_DATA_SIZE; // reset the mbuf
			m->head = m->tail - cb->hcount_key_size[RITAS_MVC_VECT];
			if(m->head < m->data) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
				return -1;
			}			
			memcpy(m->head, cb->hcount_key[RITAS_MVC_VECT], cb->hcount_key_size[RITAS_MVC_VECT]);
		}

		if(!cb->upper) {
			/*
			 * Deliver to the application
			 */
			ctx->aproxy->mvalue = m;					 
			pthread_mutex_lock(&ctx->aproxy->mutex);
			ctx->aproxy->count = 1;
			pthread_cond_broadcast(&ctx->aproxy->cond);
			pthread_mutex_unlock(&ctx->aproxy->mutex);
		}
		else {
			/*
			 * Deliver to the parent protocol
			 */
			#ifdef DEBUG
			if(m)
				printf("[%d] MVC(%d) deliver u_int value %d (%d bytes)\n", ctx->pid, cb->id, *(u_int *)m->head, m->tail - m->head);
			else
				printf("[%d] MVC(%d) deliver default value\n", ctx->pid, cb->id);
			#endif
			
			m->protocol = RITAS_MBUF_MVC;
			n = ritas_vc_input(ctx, cb->parent, m);
			if(n == -1)
				return -1;
		}
		
	}	

	return 1;
}

int ritas_mvc_output(ritas_t *ctx, ritas_mvccb_t *cb, ritas_mbuf_t *m) {
	
	ritas_mvc_header_t *mvch;
		
	/*
	 * Build MVC Header
	 */
	m->head = m->head - RITAS_MVC_HEADER_SIZE;
	if(m->head < m->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		return -1;
	}
	mvch = (ritas_mvc_header_t *)m->head;
	mvch->type = RITAS_MVC_INIT;
	
	dprintf("[%d] MVC(%d) out size %d bytes\n", ctx->pid, cb->id, m->tail - m->head);

	/*
	 * Reliably Broadcast the message.
	 */
	return ritas_rb_output(ctx, cb->init_cb[ctx->pid], m);
}


int ritas_mvc_validate(ritas_t *ctx, ritas_mvccb_t *cb, ritas_mbuf_t *m, u_char **v, u_short *v_size) {
	
	int status;
	u_char *w;
	u_short w_size;
	u_short count;
	int i, j;
	ritas_vector_t *V;

	dprintf("[%d] MVC validating VECT from %d, size %d bytes\n", ctx->pid, cb->id, m->proc->id, m->tail - m->head);
	
	if((m->tail - m->head - RITAS_MVC_HEADER_SIZE) == 0) {
		*v = NULL;
		*v_size = 0;
		return RITAS_MVC_VALID;
	}
	
	V = ritas_vector_extract(m->head + RITAS_MVC_HEADER_SIZE, m->tail - m->head - RITAS_MVC_HEADER_SIZE);
	if(!V) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to extract vector\n", __func__);
		return -1;
	}

	status = RITAS_MVC_INCOMPLETE;
	w = NULL;
	count = 0;
	for(j = 0 ; j < V->size ; j++) {
		i = *(u_short *)V->element[j];

		if(w == NULL) {
			if(cb->init_vproc_key_size[i] > 0) {
				w = cb->init_vproc_key[i];
				w_size = cb->init_vproc_key_size[i];
				dprintf("[%d] MVC w{%d}: %d (%d bytes)\n", ctx->pid, i, *(u_int*)cb->init_vproc_key[i], cb->init_vproc_key_size[i]);				
				count++;
				continue;
			}
		}

		#ifdef DEBUG
		if(cb->init_vproc_key[i])
			dprintf("[%d] MVC VECT %d: %d (%d bytes)\n", ctx->pid, i, *(u_int*)cb->init_vproc_key[i], cb->init_vproc_key_size[i]);
		else
			dprintf("[%d] MVC VECT %d: NULL (0 bytes)\n", ctx->pid, i);
		#endif

		if(cb->init_vproc_key_size[i] > 0) {
			status = RITAS_MVC_INVALID;
			if(w == cb->init_vproc_key[i]) {
				count++;
				status = RITAS_MVC_INCOMPLETE;
			}
			else
				break;		
		}
	}
	
	ritas_vector_destroy(V);
	
	if(status == RITAS_MVC_INCOMPLETE && count >= ctx->n - 2*ctx->f) {
		status = RITAS_MVC_VALID;
		*v = w;
		*v_size = w_size;
	}

	return status;
}


ritas_mvccb_t *ritas_mvc_init(ritas_t *ctx, u_short mvcid, void *parent) {
	
	int i;
	u_short txid;
	
	ritas_mvccb_t *cb;
	
	cb = malloc(sizeof(ritas_mvccb_t));
	if(!cb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() mvccb\n", __func__);
		goto bad;
	}

	/*
	 * Check for parent protocol.
	 */
	if(parent) {
		cb->upper = 1;
		cb->parent = parent;
	}
	else {
		cb->upper = 0;
		cb->parent = NULL;
		i = ritas_aproxy_init(ctx);
		if(i == -1)
			goto bad;
	}
	
	cb->init_vproc_key = NULL;
	cb->init_vproc_key_size = NULL;
	cb->init_vproc_key_total_size = 0;
	cb->init_cb = NULL;
	cb->vect_cb = NULL;
	cb->bccb = NULL;
	cb->id = mvcid;
	cb->step = 1;
	cb->bc_decision = RITAS_MVC_BC_UNDECIDED;
	
	cb->incomplete_l = ritas_list_init();
	if(!cb->incomplete_l) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init list\n", __func__);
		goto bad;
	}
	cb->incomplete = 0;
			
	for(i = 0 ; i < 2 ; i++) {
		cb->tcount[i] = 0;
		cb->hcount[i] = 0;
		cb->hcount_key[i] = NULL;
		cb->hcount_key_size[i] = 0;
		cb->vtable[i] = ritas_ht_init(RITAS_MVC_VTABLE_SIZE);
		if(!cb->vtable[i]) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init vtable\n", __func__);
			goto bad;
		}
	}
	cb->vect_dcount = 0;

	cb->init_vproc_key = malloc(ctx->n*sizeof(u_char *));
	if(!cb->init_vproc_key) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() vproc key array\n", __func__);
		goto bad;
	}
	memset(cb->init_vproc_key, '\0', ctx->n*sizeof(u_char *));
	
	cb->init_vproc_key_size = malloc(ctx->n*sizeof(u_short));
	if(!cb->init_vproc_key_size) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() vproc key size array\n", __func__);
		goto bad;
	}
	memset(cb->init_vproc_key_size, '\0', ctx->n*sizeof(u_short));

	cb->bccb = ritas_bc_init(ctx, (parent ? ((ritas_vccb_t *)parent)->id : mvcid), cb);
	if(!cb->bccb)
		goto bad;

	cb->init_cb = malloc(ctx->n*sizeof(*cb->init_cb));
	if(!cb->init_cb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() init cb\n", __func__);
		goto bad;
	}
	memset(cb->init_cb, '\0', ctx->n*sizeof(*cb->init_cb));
	

	cb->vect_cb = malloc(ctx->n*sizeof(*cb->vect_cb));
	if(!cb->vect_cb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() vect rbcb\n", __func__);
		goto bad;
	}
	memset(cb->vect_cb, '\0', ctx->n*sizeof(*cb->vect_cb));

	for(txid = 0 ; txid < ctx->n ; txid++) {
		cb->init_cb[txid] = ritas_rb_init(ctx, txid, 1, cb, RITAS_MVC, 0);	
		if(!cb->init_cb[txid])
			goto bad;
		cb->vect_cb[txid] = NULL;
	}	
	
	return cb;
	
bad:
	ritas_mvc_destroy(ctx, cb);
	return NULL;
}


void ritas_mvc_destroy(ritas_t *ctx, ritas_mvccb_t *cb) {
	
	u_short txid;
	
	if(!cb)
		return;
	
	if(!cb->upper)
		ritas_aproxy_destroy(ctx);
	
	ritas_list_destroy_mbuf(cb->incomplete_l);
	
	if(cb->init_cb) {
		for(txid = 0 ; txid < ctx->n ; txid++)
			ritas_rb_destroy(ctx, cb->init_cb[txid]);			
		free(cb->init_cb);
	}
	if(cb->vect_cb) {
		for(txid = 0 ; txid < ctx->n ; txid++)
			if(cb->vect_cb[txid])
#ifdef ECHO
				ritas_eb_destroy(ctx, cb->vect_cb[txid]);
#else
				ritas_rb_destroy(ctx, cb->vect_cb[txid]);
#endif			
		free(cb->vect_cb);
	}
	
	ritas_bc_destroy(ctx, cb->bccb);
	
	ritas_ht_destroy(cb->vtable[RITAS_MVC_INIT], NULL);
	ritas_ht_destroy(cb->vtable[RITAS_MVC_VECT], NULL);
	
	free(cb->init_vproc_key);
	free(cb->init_vproc_key_size);
	
	free(cb);
	
	return;
}

int ritas_mvc(ritas_t *ctx, u_short mvcid, u_char *prop, u_short prop_size, u_char *decision, u_short decision_size) {

	ritas_mvccb_t *cb;
	ritas_mbuf_t *mbuf;
	u_short vsize;
	int n;
	
	pthread_mutex_lock(&ctx->temp_mutex);
	
	// Init MVC control block
	cb = ritas_mvc_init(ctx, mvcid, NULL);
	if(!cb)
		goto bad;
		
	// Init mbuf
	mbuf = ritas_mbuf_init();
	if(!mbuf) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init mbuf\n", __func__);
		goto bad;
	}
	mbuf->head = mbuf->head - prop_size;
	if(mbuf->head < mbuf->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		goto bad;
	}	
	memcpy(mbuf->head, prop, prop_size);

	// Send the proposal to the MVC layer	
	n = ritas_mvc_output(ctx, cb, mbuf);
	if(n == -1)
		goto bad;
	
	pthread_mutex_unlock(&ctx->temp_mutex);
	
	// Wait until the MVC layer decides
	pthread_mutex_lock(&ctx->aproxy->mutex);
	while(!ctx->aproxy->count)
		pthread_cond_wait(&ctx->aproxy->cond, &ctx->aproxy->mutex);
	pthread_mutex_unlock(&ctx->aproxy->mutex);

	pthread_mutex_lock(&ctx->temp_mutex);

	if(!ctx->aproxy->mvalue) {
		vsize = 0;
	}
	else {
		vsize = ctx->aproxy->mvalue->tail - ctx->aproxy->mvalue->head;
		if(vsize > decision_size) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): decision buffer overflow\n", __func__);
			goto bad;
		}
		memcpy(decision, ctx->aproxy->mvalue->head, vsize);
	}
	
	ritas_mvc_destroy(ctx, cb);
	
	pthread_mutex_unlock(&ctx->temp_mutex);	
	
	return (u_int)vsize;

bad:
	if(cb) {
		ritas_mvc_destroy(ctx, cb);
		ritas_mbuf_destroy(mbuf);
	}
	return -1;
}

u_short ritas_mvc_id(ritas_mvccb_t *cb) {
	return cb->id;
}

void *ritas_mvc_parent(ritas_mvccb_t *cb) {
	return cb->parent;
}
