#ifndef RITAS_MVC_H
#define RITAS_MVC_H

#include "ritas.h"
#include "ritas_mbuf.h"
#include "ritas_ht.h"
#include "ritas_bc.h"
#include "ritas_rb.h"
#ifdef MEMWATCH
#include "memwatch.h"
#endif

#define RITAS_MVC_INIT 0
#define RITAS_MVC_VECT 1

struct ritas_mvccb {
	u_char id;
	u_char step;
	ritas_rbcb_t **init_cb;
#ifdef ECHO
	ritas_ebcb_t **vect_cb;
#else
	ritas_rbcb_t **vect_cb;
#endif
	ritas_bccb_t *bccb;
	u_char upper;
	void *parent;
#define RITAS_MVC_BC_UNDECIDED	2	
	u_char bc_decision;

	u_short tcount[2];			// total INIT/VECT messages received
	u_short hcount[2];			// highest INIT/VECT count observed for a single value
	u_char *hcount_key[2];		// INIT/VECT message key with highest count 
	u_short hcount_key_size[2];
#define RITAS_MVC_VTABLE_SIZE 7	
	ritas_ht_t *vtable[2];	
	
	u_char **init_vproc_key;	// INIT messages keys indexed by process
	u_short *init_vproc_key_size;
	u_short init_vproc_key_total_size; // TOTAL size of keys

	u_short vect_dcount;
};
typedef struct ritas_mvccb ritas_mvccb_t;


struct ritas_mvc_val {
	u_short count;
	u_short vect_count;
};
typedef struct ritas_mvc_val ritas_mvc_val_t;


struct ritas_mvc_header {
	u_short type;
	u_short value_s;
};
typedef struct ritas_mvc_header ritas_mvc_header_t;
#define RITAS_MVC_HEADER_SIZE 4

ritas_mvccb_t *ritas_mvc_init(ritas_t *ctx, u_short mvcid, void *parent);
void ritas_mvc_destroy(ritas_t *ctx, ritas_mvccb_t *cb);
int ritas_mvc_input(ritas_t *ctx, ritas_mvccb_t *cb, ritas_mbuf_t *m);
int ritas_mvc_output(ritas_t *ctx, ritas_mvccb_t *cb, ritas_mbuf_t *m);
int ritas_mvc_validate(ritas_t *ctx, ritas_mvccb_t *cb, ritas_mbuf_t *m);
u_short ritas_mvc_id(ritas_mvccb_t *cb);
void *ritas_mvc_parent(ritas_mvccb_t *cb);
int ritas_mvc(ritas_t *ctx, u_short mvcid, u_char *prop, u_short prop_size, u_char *decision, u_short decision_size);
#endif /* RITAS_MVC_H */
