#include "ritas_netio.h"
#include "ritas_rb.h"
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <assert.h>

int ritas_send(ritas_t *ctx, ritas_proc_t *proc, ritas_mbuf_t *mbuf) {

	struct timeval t0, t1;	
	int n;
	ritas_mbuf_t *m;
#ifdef TCP
	u_short len;
	struct iovec vector[2];
#endif		

	#ifdef BC_CRASH
	if(ctx->pid < ctx->f) {
		return 1;
	}
	#endif

	ritas_rb_header_t *rbh = (ritas_rb_header_t *)(mbuf->head + 2);
	
	if(proc->id == ctx->pid) {
		m = ritas_mbuf_init();
		memcpy(m->data, mbuf->head, mbuf->tail - mbuf->head);
		m->head = m->data;
		m->tail = m->head + (mbuf->tail - mbuf->head);

		pthread_mutex_lock(&ctx->scheduler_mutex);
		ritas_list_enqueue(ctx->scheduler_l, m);
		ctx->scheduler_v = 1;
		pthread_cond_broadcast(&ctx->scheduler_cond);
		pthread_mutex_unlock(&ctx->scheduler_mutex);
		return 1;
	}
	
#ifdef TCP
	len = mbuf->tail - mbuf->head;
	vector[0].iov_base = &len;
	vector[0].iov_len = sizeof(len);
	vector[1].iov_base = mbuf->head;
	vector[1].iov_len = len;
	
	n = writev(proc->socket, vector, 2);

	
/*
  	gettimeofday(&t0, NULL);
	n = send(proc->socket, &len, sizeof(len), 0);
	if(n != sizeof(len))
		printf("ALERT send(1) n(%d) != len(%d)\n", n, len);
	n = send(proc->socket, mbuf->head, len, 0);
	if(n != len)
		printf("ALERT send(2) n(%d) != len(%d)\n", n, len);	
	gettimeofday(&t1, NULL);
	dprintf("TIME writev(): %d us\n", toLL(t1)-toLL(t0));
	ctx->st += toLL(t1)-toLL(t0);		
*/
#else	

//#define toLL(tv) ((tv.tv_sec*(long long)1000000)+tv.tv_usec)	
//	gettimeofday(&t0, NULL);
	n = sendto(ctx->socket, mbuf->head, mbuf->tail - mbuf->head, 0, 
		(struct sockaddr *)&proc->addr, sizeof(proc->addr));
//	gettimeofday(&t1, NULL);
//	dprintf("TIME sendto(): %d us\n", toLL(t1)-toLL(t0));	
	ctx->st += toLL(t1)-toLL(t0);
#endif		
	if(n == -1) {
	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): send/sendto()\n", __func__);
		return -1;
	}
	
	return 1; 
}
 

