/*
 * Group management routines.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include "ritas_proc.h"

int ritas_proc_add_ipv4(ritas_t *ctx, u_short pid, u_char *ip, u_short port, u_char *key) {
	
	struct sockaddr_in addr;
	int r;
	
	memset(&addr, '\0', sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	r = inet_aton((char *)ip, &addr.sin_addr);
	if(!r) {
	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): inet_aton()\n", __func__);		
		return -1;
	}
	
	return ritas_proc_add_sockaddr_in(ctx, pid, &addr, key);
}	


int ritas_proc_add_sockaddr_in(ritas_t *ctx, u_short pid, struct sockaddr_in *addr, u_char *key) {

	int n;
#ifdef TCP
	int flag;
	struct linger lg;
	int sockbuf;
#endif

	if(pid > ctx->n-1) {
	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): pid is out of bounds\n", __func__);		
		return -1;
	}
	
	if(ctx->group[pid]) {
	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): a process with that pid already exists\n", __func__);		
		return -1;
	}
	
	ctx->group[pid] = malloc(sizeof(ritas_proc_t));
	if(!ctx->group[pid]) {
	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): malloc()\n", __func__);
	    return -1;
	}
	ctx->group[pid]->id = pid;
	memcpy(&ctx->group[pid]->addr, addr, sizeof(struct sockaddr_in));

	ctx->group[pid]->key_s = MD5_DIGEST_LENGTH;
	memset(ctx->group[pid]->key, '\0', MD5_DIGEST_LENGTH);

	memcpy(ctx->group[pid]->key, key, RITAS_KEY_SIZE);

#ifdef TCP
	/*
	 * Create socket.
	 */
	ctx->group[pid]->socket = socket(AF_INET, SOCK_STREAM, 0);
	if(ctx->group[pid]->socket < 0) {
        snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): socket()\n", __func__);
		return -1;
	}

	flag = 1;
	setsockopt(ctx->group[pid]->socket, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag));
	lg.l_onoff = 1;
	lg.l_linger = 0;
	setsockopt(ctx->group[pid]->socket, SOL_SOCKET, SO_LINGER, (char *)&lg, sizeof(lg));
	sockbuf = 1024*1024;
	setsockopt(ctx->group[pid]->socket, SOL_SOCKET, SO_RCVBUF, &sockbuf, sizeof(sockbuf));
	setsockopt(ctx->group[pid]->socket, SOL_SOCKET, SO_SNDBUF, &sockbuf, sizeof(sockbuf));
#endif

	return 1;
}


ritas_proc_t *ritas_proc_get(ritas_t *ctx, u_short pid) {
	
	if(pid > ctx->n-1) {
	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): pid is out of bounds\n", __func__);		
		return NULL;
	}
		
	return ctx->group[pid];
}

int ritas_group_size(ritas_t *ctx) {
	return ctx->n;
}
