#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include "ritas.h"
#include "ritas_rb.h"
#include "ritas_rc.h"
#include "ritas_bc.h"
#include "ritas_mvc.h"
#include "ritas_vc.h"
#include "ritas_ab.h"

/*
 * TODO: WHAT IF THIS NEVER ENDS?
 */

int ritas_rb_input(ritas_t *ctx, ritas_rbcb_t *rbcb, ritas_mbuf_t *m) {

	ritas_rb_header_t *rbh;
	u_char *key, ksize;
	u_short *vcount;
	int i;
	u_short type;

	if(rbcb->step > 3) {
		ritas_mbuf_destroy(m);
		return 1;
	}
	rbh = (ritas_rb_header_t *)m->head;

	rbprintf("%ld [%d] RB in type: %d from %d\n", ritas_get_timer(ctx), ctx->pid, rbh->type, m->proc->id);

	//#ifdef DEBUG
	//bch = m->head + RITAS_RB_HEADER_SIZE;
	//printf("DIN RC(%d,0,{%d,%d}) RB(%d) BC(%d,%d,%d)\n", ((ritas_bccb_t *)rbcb->parent)->id, rbcb->txid, rbcb->rbid, rbh->type, bch->round, bch->step, bch->value);
	//#endif

	/*
	 * Check if this message is still relevant for the step we're in.
	 * We're going to be pedantic for the sake of sanitization.
	 */
	switch(rbcb->step) {
		case 2:
			if(rbh->type != RITAS_RB_ECHO && rbh->type != RITAS_RB_READY) {
				ritas_mbuf_destroy(m);
				return 1;
			}
			break;
		case 3:
			if(rbh->type != RITAS_RB_READY) {
				ritas_mbuf_destroy(m);
				return 1;
			}
			break;
	}

	/*
	 * Check if we already have a message from this process.
	 */
	if(rbcb->vproc[rbh->type][m->proc->id]) {
		ritas_mbuf_destroy(m);
		return 1;
	}

	/*
	 * Locate the info holder for the value received
	 */
	key = m->head + RITAS_RB_HEADER_SIZE;
	ksize = m->tail - key;
	vcount = ritas_ht_lookup(rbcb->vtable[rbh->type], key, ksize);
	if(!vcount) {
		/*
		 * It's a new value. Create a new holder for it.
		 */
		vcount = malloc(sizeof(*vcount));
		if(!vcount) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): malloc()\n", __func__);
			return -1;
		}
		*vcount = 0;
		i = ritas_ht_add(rbcb->vtable[rbh->type], key, ksize, vcount);
		if(i == -1) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to insert object in hash table\n", __func__);			
			return -1;
		}
	}

	/*
	 * Update the count.
	 */
	(*vcount)++;
	rbcb->vproc[rbh->type][m->proc->id] = 1;

	/*
	 * Take the apropriate action according to the new state.
	 */
	if(*vcount >= rbcb->threshold[rbh->type]) {
		rbprintf("%ld us [%d] RB threshold step %d\n", ritas_get_timer(ctx), ctx->pid, rbcb->step);
		switch(rbcb->step) {
			case 1:
			type = rbh->type;
			rbh->type = RITAS_RB_ECHO; // reuse received mbuf. HEH.
			i = m->tail - m->head;
			m->flags |= RITAS_MBUF_PROTECTED;
			i = ritas_rc_broadcast(ctx, rbcb->rccb, m);
			if(i == -1)
				return -1;
			rbcb->step++;
			if(type == RITAS_RB_INITIAL) {
				m->flags ^= RITAS_MBUF_PROTECTED;
				ritas_mbuf_destroy(m);
				break;
			}
			m->head = (u_char *)rbh;


			case 2:
			if(!(rbcb->flags & RITAS_RB_FLG_ECHO)) {
				rbh->type = RITAS_RB_READY; // reuse received mbuf. HEH.
				i = m->tail - m->head;
				if(m->flags & RITAS_MBUF_PROTECTED)
					m->flags ^= RITAS_MBUF_PROTECTED;
				i = ritas_rc_broadcast(ctx, rbcb->rccb, m);
				if(i == -1)
					return -1;
				rbcb->threshold[RITAS_RB_READY] = 2*ctx->f + 1;
				rbcb->step++;
				break;
			}
			else {
				rbcb->step++;
			}
			case 3:
			rbcb->step++;
			m->head = m->head + RITAS_RB_HEADER_SIZE;
			m->proc = ritas_proc_get(ctx, rbcb->txid);
			m->rbcb = rbcb;

			
			if(!rbcb->upper) {
				/*
				 * Deliver to the application
				 */
				ctx->aproxy->mvalue = m;
				pthread_mutex_lock(&ctx->aproxy->mutex);
				ctx->aproxy->count++;
				pthread_cond_broadcast(&ctx->aproxy->cond);
				pthread_mutex_unlock(&ctx->aproxy->mutex);
			}
			else {
				/*
				 * Pass to the upper layer
				 */
				m->protocol = RITAS_MBUF_RB;
				switch(rbcb->upper) {
				case RITAS_RB_UPPER_BC:
					//dprintf("[%d] RB deliver BC\n", ctx->pid);
					i = ritas_bc_input(ctx, (ritas_bccb_t *)rbcb->parent, m);
					break;
				case RITAS_RB_UPPER_MVC:
					//dprintf("[%d] RB deliver MVC\n", ctx->pid);
					i = ritas_mvc_input(ctx, (ritas_mvccb_t *)rbcb->parent, m);
					break;
				case RITAS_RB_UPPER_VC:
					//dprintf("[%d] RB deliver VC\n", ctx->pid);
					i = ritas_vc_input(ctx, (ritas_vccb_t *)rbcb->parent, m);
					break;
				case RITAS_RB_UPPER_AB:
					//dprintf("[%d] RB deliver AB\n", ctx->pid);
					i = ritas_ab_input(ctx, (ritas_abcb_t *)rbcb->parent, m);
					break;
				}
				if (i == -1)
					return -1;

			}
			break;
		}
	}
	else {
		ritas_mbuf_destroy(m);
	}
		
	return 1;
}


int ritas_rb_output(ritas_t *ctx, ritas_rbcb_t *rbcb, ritas_mbuf_t *m) {
	
	ritas_rb_header_t *rbh;
	
	/*
	 * Build RB Header
	 */
	m->head = m->head - RITAS_RB_HEADER_SIZE;
	if(m->head < m->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		return -1;
	}
	rbh = (ritas_rb_header_t *)m->head;
	rbh->type = RITAS_RB_INITIAL;

	//dprintf("[%d] RB out type: %d in\n", ctx->pid, rbh->type);
	
	/*
	 * Broadcast the message over the reliable channel
	 */
	return ritas_rc_broadcast(ctx, rbcb->rccb, m);
}


ritas_rbcb_t *ritas_rb_init(ritas_t *ctx, u_short txid, u_short rbid, void *parent, u_char parent_type, int flags) {
	
	ritas_rbcb_t *rbcb;
	int i;
	
	/*
	 * Create RBCB
	 */
	rbcb = (ritas_rbcb_t *)malloc(sizeof(ritas_rbcb_t));
	if(!rbcb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() control block\n", __func__);
		goto bad;
	}

	rbcb->flags = flags;

	/*
	 * Check for parent protocol.
	 */
	if(parent) {
		if(parent_type < 1 || parent_type > 4) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): invalid parent type\n", __func__);
			goto bad;
		}
		rbcb->upper = parent_type;
		rbcb->parent = parent;	
	}
	else {
		rbcb->upper = 0;
		rbcb->parent = NULL;
		i = ritas_aproxy_init(ctx);
		if(i == -1)
			goto bad;
	}


	rbcb->txid = txid;
	rbcb->rbid = rbid;
	/*
	 * Build RITAS ID
	 */
	//ritas_rb_build_id(rbcb, txid, rbid);
	
	/*
	 * Init thresholds and step and list pointers.
	 */
	rbcb->threshold[RITAS_RB_INITIAL] = 1;
	rbcb->threshold[RITAS_RB_ECHO] = ctx->n - ctx->f;
	rbcb->threshold[RITAS_RB_READY] = ctx->f + 1;
	rbcb->step = 1;

	/*
	 * Init value table for each step.
	 */
	for(i = 0 ; i < 3 ; i++) {
		rbcb->vproc[i] = malloc(sizeof(u_char)*ctx->n);
		if(!rbcb->vproc[i]) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() vproc array\n", __func__);
			goto bad;
		}
		memset(rbcb->vproc[i], '\0', sizeof(u_char)*ctx->n);
		rbcb->vtable[i] = ritas_ht_init(RITAS_RB_VTABLE_SIZE);
		if(!rbcb->vtable[i]) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init hash table\n", __func__);
			goto bad;
		}
	}

	rbcb->rccb = ritas_rc_init(ctx, rbcb, RITAS_RC_PROTO_RB);
	if(!rbcb->rccb) {
		goto bad;
	}
		
	/*
	 * Add to OOC list.
	 */
/*
	i = ritas_list_enqueue(ctx->l_ooc, rbcb);
	if(i == -1) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to list\n", __func__);
		goto bad;
	}
	pthread_cond_broadcast(&ctx->scheduler_cond);
*/
	/*
	 * Insert RBCB in the hash table
	 */
/*	i = ritas_ht_add(ctx->ht_rbcb, (u_char *)&rbcb->id, RITAS_ID_HEADER_SIZE, rbcb);
	if(i == -1)
		goto bad;
*/		
	return rbcb;
	
bad:
	if(rbcb) {
		if(!rbcb->upper)
			ritas_aproxy_destroy(ctx);
		for(i = 0 ; i < 3 ; i++) {
			free(rbcb->vproc[i]);
			ritas_ht_destroy(rbcb->vtable[i], NULL);
		}
		free(rbcb);
	}
	return NULL;
}

void ritas_rb_destroy(ritas_t *ctx, ritas_rbcb_t *rbcb) {
	
	int i;

	if(rbcb) {	
		/*
		 * Remove RBCB from the hash table.
		 */
//		ritas_ht_remove(ctx->ht_rbcb, (u_char *)&rbcb->id, RITAS_ID_HEADER_SIZE);

		/*
		 * Remove RBCB from OOC list (if it's there in the first place)
		 */
//		if(rbcb->prev || rbcb->next) {
//			ritas_list_remove(ctx->l_ooc, rbcb);
//		}

		/*
		 * Consume all OOC messages.
		 */
//		if((mfifo = ritas_ht_lookup(ctx->ht_ooc, (u_char *)&rbcb->id, RITAS_ID_HEADER_SIZE)))
//			while((m = ritas_list_dequeue(mfifo)))
//				ritas_mbuf_destroy(m);

		/*
		 * Destroy aproxy
		 */
		if(!rbcb->upper)
			ritas_aproxy_destroy(ctx);

		/*
		 * Destroy value tables.
		 */
		for(i = 0 ; i < 3 ; i++) {
			free(rbcb->vproc[i]);
			ritas_ht_destroy(rbcb->vtable[i], NULL);
		}
	
		ritas_rc_destroy(ctx, rbcb->rccb);
	
		free(rbcb);
	}
	
	return;
}

int ritas_rb_renew(ritas_t *ctx, ritas_rbcb_t *rbcb, u_short txid, u_short rbid) {

	int i;
	
	if(!rbcb) {
		return 1;
	}
	
	/*
	 * Remove RBCB from the hash table
	 */
	//ritas_ht_remove(ctx->ht_rbcb, (u_char *)&rbcb->id, RITAS_ID_HEADER_SIZE);

	/*
	 * Remove RBCB from OOC list (if it's there in the first place)
	 */
	//if(rbcb->prev || rbcb->next) {
	//	ritas_list_remove(ctx->l_ooc, rbcb);
	//}

	/*
	 * Consume all OOC messages.
	 */
	//mfifo = ritas_ht_lookup(ctx->ht_ooc, (u_char *)&rbcb->id, RITAS_ID_HEADER_SIZE);
	//if(mfifo)
	//	while((m = ritas_list_dequeue(mfifo)))
	//		ritas_mbuf_destroy(m);

	/*
	 * Build RITAS ID
	 */
	//ritas_rb_build_id(rbcb, txid, rbid);

	/*
	 * Renew thresholds and step.
	 */
	rbcb->threshold[RITAS_RB_INITIAL] = 1;
	rbcb->threshold[RITAS_RB_ECHO] = ctx->n - ctx->f;
	rbcb->threshold[RITAS_RB_READY] = ctx->f + 1;
	rbcb->step = 1;
	rbcb->txid = txid;
	rbcb->rbid = rbid;


	/*
	 * Renew value table for each step.
	 */
	for(i = 0 ; i < 3 ; i++) {
		memset(rbcb->vproc[i], '\0', sizeof(u_char)*ctx->n);
		/*
		 * TODO: these two steps could be made more efficient.
		 */
		ritas_ht_destroy(rbcb->vtable[i], NULL);
		rbcb->vtable[i] = ritas_ht_init(RITAS_RB_VTABLE_SIZE);
		if(!rbcb->vtable[i]) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init hash table\n", __func__);
			return -1;
		}
	}

	i = ritas_rc_renew(ctx, rbcb->rccb, rbcb);
	if(i == -1) {
		return -1;
	}
	
	/*
	 * Insert RBCB in the hash table
	 */
	//i = ritas_ht_add(ctx->ht_rbcb, (u_char *)&rbcb->id, RITAS_ID_HEADER_SIZE, rbcb);
	//if(!i) {
	//	snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to hash table\n", __func__);
	//	return -1;
	//}		

	/*
	 * Add to OOC list.
	 */
	//i = ritas_list_enqueue(ctx->l_ooc, rbcb);
	//if(i == -1) {
	//	snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to list\n", __func__);
	//	return -1;
	//}
	//pthread_cond_broadcast(&ctx->scheduler_cond);

	return 1;

}

int ritas_rb_bcast(ritas_t *ctx, u_short rbid, u_char *buf, u_short buf_s) {

	ritas_rbcb_t *rbcb;
	ritas_mbuf_t *mbuf;

	pthread_mutex_lock(&ctx->temp_mutex);

	rbcb = ritas_rb_init(ctx, ctx->pid, rbid, NULL, 0, 0);
	if(!rbcb) {
		goto bad;
	}
	
	mbuf = ritas_mbuf_init();
	mbuf->head = mbuf->head - buf_s;
	memcpy(mbuf->head, buf, buf_s);

	ritas_rb_output(ctx, rbcb, mbuf);

	pthread_mutex_unlock(&ctx->temp_mutex);

	pthread_mutex_lock(&ctx->aproxy->mutex);
	while(ctx->aproxy->count < 1)
		pthread_cond_wait(&ctx->aproxy->cond, &ctx->aproxy->mutex);
	pthread_mutex_unlock(&ctx->aproxy->mutex);

	pthread_mutex_lock(&ctx->temp_mutex);

	if(!ctx->aproxy->mvalue)
		goto bad;
		
	ritas_rb_destroy(ctx, rbcb);

	pthread_mutex_unlock(&ctx->temp_mutex);
	
	return 1;
	
bad:
	if(rbcb) {
		ritas_rb_destroy(ctx, rbcb);
	}
	return -1;
}	

int ritas_rb_recv(ritas_t *ctx, u_short txid, u_short rbid, u_char *buf, u_short buf_s) {

	ritas_rbcb_t *rbcb;
	u_short mvalue_s;

	pthread_mutex_lock(&ctx->temp_mutex);

	rbcb = ritas_rb_init(ctx, txid, rbid, NULL, 0, 0);
	if(!rbcb)
		goto bad;
	
	
	pthread_mutex_lock(&ctx->scheduler_mutex);
	ctx->scheduler_v = 1;
	pthread_cond_broadcast(&ctx->scheduler_cond);
	pthread_mutex_unlock(&ctx->scheduler_mutex);
	
	pthread_mutex_unlock(&ctx->temp_mutex);
	
	pthread_mutex_lock(&ctx->aproxy->mutex);
	while(ctx->aproxy->count < 1)
		pthread_cond_wait(&ctx->aproxy->cond, &ctx->aproxy->mutex);
	pthread_mutex_unlock(&ctx->aproxy->mutex);

	pthread_mutex_lock(&ctx->temp_mutex);

	if(!ctx->aproxy->mvalue)
		goto bad;

	mvalue_s = ctx->aproxy->mvalue->tail - ctx->aproxy->mvalue->head;
	if(mvalue_s > buf_s) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): buffer overflow\n", __func__);
		goto bad;
	}

	memcpy(buf, ctx->aproxy->mvalue->head, mvalue_s);
	
	ritas_rb_destroy(ctx, rbcb);

	pthread_mutex_unlock(&ctx->temp_mutex);
		
	return mvalue_s;
	
bad:
	if(rbcb) {
		ritas_rb_destroy(ctx, rbcb);
	}
	return -1;
}

u_short ritas_rb_txid(ritas_rbcb_t *rbcb) {
	return rbcb->txid;
}

u_short ritas_rb_rbid(ritas_rbcb_t *rbcb) {
	return rbcb->rbid;
}

void *ritas_rb_parent(ritas_rbcb_t *rbcb) {
	return rbcb->parent;
}

u_char ritas_rb_parent_type(ritas_rbcb_t *rbcb) {
	return rbcb->upper;
}

/*
void ritas_rb_build_id(ritas_rbcb_t *rbcb, u_short txid, u_short rbid) {

	u_char upper;
	void *cb;

	rbcb->id.flags = 0x00;
	rbcb->id.mvcid = 0;
	rbcb->id.cid = 0;
	rbcb->id.txid = txid;
	rbcb->id.rbid = rbid;
	upper = rbcb->upper;
	cb = rbcb->parent;

	switch(upper) {
	case RITAS_RB_UPPER_BC:
		rbcb->id.cid = ((ritas_bccb_t *)cb)->id;
		rbcb->id.flags |= RITAS_ID_BC;
		upper = ((ritas_bccb_t *)cb)->upper;
		if(!upper)
			break;
		cb = ((ritas_bccb_t *)cb)->parent;
	
	case RITAS_RB_UPPER_MVC:
		rbcb->id.mvcid = ((ritas_mvccb_t *)cb)->id;
		rbcb->id.flags |= RITAS_ID_MVC;
		upper = ((ritas_mvccb_t *)cb)->upper;
		if(!upper)
			break;
		cb = ((ritas_mvccb_t *)cb)->parent;		
	case RITAS_RB_UPPER_VC:
		if(!rbcb->id.cid)
			rbcb->id.cid = ((ritas_vccb_t *)cb)->id;
		rbcb->id.flags |= RITAS_ID_VC;
		upper = ((ritas_vccb_t *)cb)->upper;
		if(!upper)
			break;
		cb = ((ritas_vccb_t *)cb)->parent;	
	case RITAS_RB_UPPER_AB:
		if(!rbcb->id.cid)
			rbcb->id.cid = ((ritas_abcb_t *)cb)->id;
		rbcb->id.flags |= RITAS_ID_AB;
	}

}
*/
