#ifndef RITAS_RB_H
#define RITAS_RB_H

#include <pthread.h>
#include "ritas_rc.h"
#include "ritas_mbuf.h"
#include "ritas_ht.h"
#include "ritas_proc.h"
#include "ritas_aproxy.h"
#ifdef MEMWATCH
#include "memwatch.h"
#endif

#define RITAS_RB_HASH(rbid)	(rbid % RITAS_RBCB_SIZE)

struct ritas_rbcb {
	u_short txid;
	u_short rbid;
#define RITAS_RB_FLG_ECHO 0x01
	int flags;
	u_short threshold[3];
	u_char step;
#define RITAS_RB_UPPER_BC	1
#define RITAS_RB_UPPER_MVC	2
#define RITAS_RB_UPPER_VC	3
#define RITAS_RB_UPPER_AB	4	
	u_char upper;
	void *parent;
	u_char *vproc[3];
#define RITAS_RB_VTABLE_SIZE 7
	ritas_ht_t *vtable[3];
	ritas_rccb_t *rccb;
};
typedef struct ritas_rbcb ritas_rbcb_t;


struct ritas_rb_header {
	u_short type;
#define RITAS_RB_INITIAL	0x00
#define RITAS_RB_ECHO		0x01
#define RITAS_RB_READY		0x02
}; 
typedef struct ritas_rb_header ritas_rb_header_t;
#define RITAS_RB_HEADER_SIZE 2

ritas_rbcb_t *ritas_rb_init(ritas_t *ctx, u_short txid, u_short id, void *parent, u_char parent_type, int flags);
void ritas_rb_destroy(ritas_t *ctx, ritas_rbcb_t *rbcb);
int ritas_rb_renew(ritas_t *ctx, ritas_rbcb_t *rbcb, u_short txid, u_short rbid);
int ritas_rb_input(ritas_t *ctx, ritas_rbcb_t *rbcb, ritas_mbuf_t *m);
int ritas_rb_output(ritas_t *ctx, ritas_rbcb_t *rbcb, ritas_mbuf_t *m);
int ritas_rb_bcast(ritas_t *ctx, u_short rbid, u_char *buf, u_short buf_s);
int ritas_rb_recv(ritas_t *ctx, u_short txid, u_short rbid, u_char *buf, u_short buf_s);
//void ritas_rb_build_id(ritas_rbcb_t *rbcb, u_short txid, u_short rbid);
u_short ritas_rb_txid(ritas_rbcb_t *rbcb);
u_short ritas_rb_rbid(ritas_rbcb_t *rbcb);
void *ritas_rb_parent(ritas_rbcb_t *rbcb);
u_char ritas_rb_parent_type(ritas_rbcb_t *rbcb);
#endif /* RITAS_RB_H */
