

ritas_rbcb_t *ritas_rb_init(ritas_t *ctx, u_short txid, u_short id, void *parent, u_char parent_type) {
	ritas_rbcb_t *rbcb;
	
	switch(ctx->protocol_rb) {
		case RITAS_RB_BRACHA:
		rbcb = ritas_rb_bracha_init(ctx, txid, id, parent, parent_type);

		default:
		return NULL;
	}

	if(parent) {
		if(parent_type < 1 || parent_type > 4) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): invalid parent type\n", __func__);
			goto bad;
		}
	}
	else {
		n = ritas_aproxy_init(ctx);
		if(n == -1)
			goto bad;
	}
	
}

void ritas_rb_destroy(ritas_t *ctx, ritas_rbcb_t *rbcb) {
	switch(ctx->protocol_rb) {
		case RITAS_RB_BRACHA:
		return ritas_rb_bracha_destroy(ctx, rbcb);

		default:
		return;
	}
	
	ritas_aproxy_destroy(ctx);
}

int ritas_rb_renew(ritas_t *ctx, ritas_rbcb_t *rbcb, u_short txid, u_short rbid) {
	switch(ctx->protocol_rb) {
		case RITAS_RB_BRACHA:
		return ritas_rb_bracha_renew(ctx, rbcb, txid, rbid);

		default:
		return;
	}
}
	
int ritas_rb_input(ritas_t *ctx, ritas_rbcb_t *rbcb, ritas_mbuf_t *m) {
	switch(ctx->protocol_rb) {
		case RITAS_RB_BRACHA:
		return ritas_rb_bracha_input(ctx, rbcb, m);

		default:
		return;
	}
}
	
int ritas_rb_output(ritas_t *ctx, ritas_rbcb_t *rbcb, ritas_mbuf_t *m) {
	switch(ctx->protocol_rb) {
		case RITAS_RB_BRACHA:
		return ritas_rb_bracha_output(ctx, rbcb, m);

		default:
		return;
	}
}

int ritas_rb_bcast(ritas_t *ctx, u_short rbid, u_char *buf, u_short buf_s) {

	ritas_rbcb_t *rbcb;
	ritas_mbuf_t *mbuf;
	ritas_aproxy_t *aproxy;

	rbcb = ritas_rb_init(ctx, ctx->pid, rbid, NULL, 0);
	if(!rbcb) {
		goto bad;
	}
	
	mbuf = ritas_mbuf_init();
	mbuf->head = mbuf->head - buf_s;
	memcpy(mbuf->head, buf, buf_s);

	pthread_mutex_lock(&ctx->aproxy->mutex);

	ritas_rb_output(ctx, rbcb, mbuf);

	while(ctx->aproxy->count < 1)
		pthread_cond_wait(&ctx->aproxy->cond, &ctx->aproxy->mutex);
	pthread_mutex_unlock(&ctx->aproxy->mutex);

	if(!ctx->aproxy->mvalue)
		goto bad;
		
	ritas_rb_destroy(ctx, rbcb);
	
	return 1;
	
bad:
	if(rbcb) {
		ritas_rb_destroy(ctx, rbcb);
	}
	return -1;
}

int ritas_rb_recv(ritas_t *ctx, u_short txid, u_short rbid, u_char *buf, u_short buf_s) {

	ritas_rbcb_t *rbcb;
	u_short mvalue_s;


	rbcb = ritas_rb_init(ctx, txid, rbid, NULL, 0);
	if(!rbcb)
		goto bad;
	
	pthread_mutex_lock(&ctx->scheduler_mutex);
	pthread_cond_broadcast(&ctx->scheduler_cond);
	pthread_mutex_unlock(&ctx->scheduler_mutex);

	pthread_mutex_lock(&ctx->aproxy->mutex);
	
	while(ctx->aproxy->count < 1)
		pthread_cond_wait(&ctx->aproxy->cond, &ctx->aproxy->mutex);
	pthread_mutex_unlock(&ctx->aproxy->mutex);

	if(!ctx->aproxy->mvalue)
		goto bad;

	mvalue_s = ctx->aproxy->mvalue->tail - ctx->aproxy->mvalue->head;
	if(mvalue_s > buf_s) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): buffer overflow\n", __func__);
		goto bad;
	}

	memcpy(buf, ctx->aproxy->mvalue->head, mvalue_s);
	
	ritas_mbuf_destroy(ctx->aproxy->mvalue);
	ritas_aproxy_destroy(ctx);
	ritas_rb_destroy(ctx, rbcb);
		
	return mvalue_s;
	
bad:
	if(rbcb) {
		ritas_rb_destroy(ctx, rbcb);
	}
	return -1;
}
