// This file is a stub. Not the real rc implementation
#ifndef RITAS_RC_H
#define RITAS_RC_H

#include <sys/types.h>
#include <pthread.h>
#include "ritas.h"
#include "ritas_mbuf.h"
#include "ritas_proc.h"
#ifdef MEMWATCH
#include "memwatch.h"
#endif

struct ritas_id_header {
#define RITAS_ID_BC 	0x01
#define RITAS_ID_MVC	0x02
#define RITAS_ID_VC		0x04
#define RITAS_ID_AB		0x08
	u_char flags;
	u_char mvcid;
	u_short cid;
	u_short txid;
	u_short rbid;
};
typedef struct ritas_id_header ritas_id_t;
#define RITAS_ID_HEADER_SIZE 8

struct ritas_rccb {
	struct ritas_rccb *prev;
	struct ritas_rccb *next;
	ritas_id_t id;
	void *parent;
#define RITAS_RC_PROTO_RB 0
#define RITAS_RC_PROTO_EB 1
#define RITAS_RC_PROTO_BC 2
	u_char parent_proto;
};
typedef struct ritas_rccb ritas_rccb_t;

struct ritas_rc_header {
	ritas_id_t id;
	u_short pid;
};
typedef struct ritas_rc_header ritas_rc_header_t;
#define RITAS_RC_HEADER_SIZE (RITAS_ID_HEADER_SIZE + 2)

ritas_rccb_t *ritas_rc_init(ritas_t *ctx, void *parent, u_char parent_proto);
void ritas_rc_destroy(ritas_t *ctx, ritas_rccb_t *rccb);
int ritas_rc_renew(ritas_t *ctx, ritas_rccb_t *rccb, void *parent);
int ritas_rc_send(ritas_t *ctx, ritas_rccb_t *rccb, ritas_mbuf_t *m, ritas_proc_t *proc);
int ritas_rc_broadcast(ritas_t *ctx, ritas_rccb_t *rccb, ritas_mbuf_t *m);
int ritas_rc_output(ritas_t *ctx, ritas_rccb_t *rccb, ritas_mbuf_t *m, ritas_proc_t *proc);
int ritas_rc_input(ritas_t *ctx, ritas_mbuf_t *m);
void ritas_rc_build_id(ritas_t *ctx, ritas_rccb_t *rccb);

struct ritas_rbto {
	struct ritas_rbto *prev;
	struct ritas_rbto *next;
	ritas_mbuf_t *m;
	u_char *proc; // array of size n
	long long tx_time;
};
typedef struct ritas_rbto ritas_rbto_t;

/*
 * Timeout function. Blocks on sigwait() waiting for the current RTO to expire
 * via SIGALRM and calls retransmission routine.
 */
void ritas_rc_timeout(ritas_t *ctx);

/*
 * Reset timer. Sets the next timeout based on the current RTO value and the
 * first RBTO on the list.
 */
void ritas_rc_reset_timer(ritas_t *ctx);
	
/*
 * Retransmission function. Iterates over the RBTO list and retransmits the
 * messages hold by the RBTOs whose RTO has expired until it finds one that
 * hasn't yet expired. The RBTOs are naturally ordered by their RTO value.
 */
int ritas_rc_rtx(ritas_t *ctx);

/*
 * RTO update function. Updates the RTO with the last measured Step Time and
 * calls the retransmission routine since the new RTO may be smaller than the
 * previous one.
 */
void ritas_rc_update_rto(ritas_t *ctx, struct timeval *measuredST);

/*
 * Calculates the time elapsed since a given moment t0 to the present moment.
 */
void ritas_rc_elapsed(long long t0, struct timeval *elapsed);

/*
 * Inits a new RBTO.
 */
ritas_rbto_t *ritas_rbto_init(ritas_t *ctx, ritas_mbuf_t *m);

/*
 * Destroys an RBTO.
 */
void ritas_rbto_destroy(ritas_rbto_t *rbto); 

/*
 * Enqueues an RBTO to the RBTO list.
 */
int ritas_rbto_list_add(ritas_t *ctx, ritas_rbto_t *rbto);

/*
 * Removes an RBTO from the RBTO list.
 */
int ritas_rbto_list_remove(ritas_t *ctx, ritas_rbto_t *rbto);

#endif /* RITAS_RC_H */
