#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ritas_vc.h"
#include "ritas_vector.h"
#include "ritas_rb.h"
#include "ritas_mvc.h"
#include "ritas_ab.h"

int ritas_vc_input(ritas_t *ctx, ritas_vccb_t *cb, ritas_mbuf_t *m) {

	ritas_mbuf_t *mbuf;
	int i, j, n;
	ritas_vector_t *vector, *X;
	ritas_vc_vproc_t *vproc;

	if(m->protocol == RITAS_MBUF_RB) {
		cb->W[m->proc->id] = m;
		cb->count++;		
		vcprintf("%ld us [%d] VC(%d) incoming RB for round %d from process %d (%d bytes)\n", ritas_get_timer(ctx), ctx->pid, cb->id, cb->round, m->proc->id, m->tail-m->head);
	}
	else if(m->protocol == RITAS_MBUF_MVC) {
		vcprintf("%ld us [%d] VC(%d) decision MVC for round %d with value size of %d bytes\n", ritas_get_timer(ctx), ctx->pid, cb->id, cb->round, m->tail - m->head);
		if(m->tail - m->head > 0) {
			if(!cb->upper) {
				/*
				 * Deliver to the application
				 */
				ctx->aproxy->mvalue = m;					 
				pthread_mutex_lock(&ctx->aproxy->mutex);
				ctx->aproxy->count = 1;
				pthread_cond_broadcast(&ctx->aproxy->cond);
				pthread_mutex_unlock(&ctx->aproxy->mutex);
			}
			else {
				/*
				 * Deliver to the parent protocol
				 */
				m->protocol = RITAS_MBUF_VC;
				ritas_ab_input(ctx, cb->parent, m);
			}
			return 1;
		}
		else {
			ritas_mbuf_destroy(m);
			cb->round++;
			ritas_mvc_destroy(ctx, cb->mvccb);
			cb->mvccb = ritas_mvc_init(ctx, cb->round+1, cb);
			if(!cb->mvccb)
				return -1;
			cb->lock = 0;
		}
	}
	else {
		ritas_mbuf_destroy(m);
		return 1;
	}

	if(!cb->lock && cb->count >= ctx->n - ctx->f + cb->round) {

		if(cb->upper) {
			vector = ritas_vector_init(ctx->n*RITAS_AB_THROUGHPUT);
			if(!vector) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't init vector\n", __func__);
				return -1;
			}		

			for(i = 0 ; i < ctx->n ; i++) {
				if(cb->W[i]) {
					X = ritas_vector_extract(cb->W[i]->head, cb->W[i]->tail - cb->W[i]->head);
					dprintf("[%d] VC ID count (%d elements): ", ctx->pid, X->size);
					for(j = 0 ; j < X->size ; j++) {
						if(X->element_s[j] > 0) {
							/*
							 * Count the number of times this given hash appeared.
							 */
							vproc = ritas_ht_lookup(cb->v_count, X->element[j], X->element_s[j]);
							if(!vproc) {
								vproc = ritas_vc_vproc_init(ctx->n);
								if(!vproc) {
									snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init vproc\n", __func__);
									return -1;
								}
							
								n = ritas_ht_add(cb->v_count, X->element[j], X->element_s[j], vproc);
								if(n == -1) {
									snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add element to hash table\n", __func__);
									return -1;
								}
							}	
							if(!vproc->proc[i]) {
								vproc->count++;
								vproc->proc[i] = 1;
							}
					
							dprintf("(%d,%d)[%d]", *(u_short *)X->element[j], *(u_short *)(X->element[j]+2), vproc->count);
							
							if(vproc->count == ctx->f+1) {
								dprintf("*");
								ritas_vector_add(vector, X->element[j], X->element_s[j]);
							}
							dprintf(" ");
						}
					}
					dprintf("\n");					
					ritas_vector_destroy(X);
				}
			}
		}
		else {
			vector = ritas_vector_init(ctx->n);
			if(!vector) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't init vector\n", __func__);
				return -1;
			}			
			
			for(i = 0 ; i < ctx->n ; i++) {
				n = ritas_vector_add(vector, (cb->W[i] ? cb->W[i]->head : NULL), (cb->W[i] ? cb->W[i]->tail - cb->W[i]->head : 0));
				if(n == -1) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): couldn't add element to vector\n", __func__);
					return -1;
				}
			}
		}
		
		mbuf = ritas_mbuf_init();
		mbuf->head = mbuf->head - vector->data_s;
		if(mbuf->head < mbuf->data) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
			return -1;
		}					
		
		#ifdef EVIL
		if(ctx->pid == 0)
		memset(mbuf->head, 'A', vector->data_s);
		else {
		#endif
		ritas_vector_build(vector, mbuf->head);
		#ifdef EVIL
		}
		#endif
		ritas_vector_destroy(vector);
				
		cb->lock = 1;
		n = ritas_mvc_output(ctx, cb->mvccb, mbuf);
		if(n == -1)
			return -1;
	}
	
	return 1;
}


int ritas_vc_output(ritas_t *ctx, ritas_vccb_t *cb, ritas_mbuf_t *m) {

		ritas_rb_output(ctx, cb->rbcb[ctx->pid], m);
		return 1;
}


ritas_vccb_t *ritas_vc_init(ritas_t *ctx, u_short cid, void *parent) {

	ritas_vccb_t *cb;
	int i;
	
	cb = malloc(sizeof(ritas_vccb_t));
	if(!cb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() vccb\n", __func__);
		goto bad;
	}

	/*
	 * Check for parent protocol.
	 */
	if(parent) {
		cb->upper = 1;
		cb->parent = parent;
	}
	else {
		cb->upper = 0;
		cb->parent = NULL;
		i = ritas_aproxy_init(ctx);
		if(i == -1)
			goto bad;
	}

	cb->id = cid;
	cb->round = 0;
	cb->lock = 0;

	cb->W = malloc(ctx->n*sizeof(ritas_mbuf_t *));
	if(!cb->W) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() W\n", __func__);
		goto bad;
	}
	memset(cb->W, '\0', ctx->n*sizeof(ritas_mbuf_t *));

	cb->count = 0;

	/*
	 * Init value count hash table.
	 */
	cb->v_count = ritas_ht_init(RITAS_VC_V_COUNT_SIZE);
	if(!cb->v_count) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init hash table\n", __func__);
		goto bad;
	}

	cb->mvccb = ritas_mvc_init(ctx, cb->round+1, cb);
	if(!cb->mvccb)
		goto bad;

	cb->rbcb = malloc(ctx->n*sizeof(ritas_rbcb_t *));
	if(!cb->rbcb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() rbcb array\n", __func__);
		goto bad;
	}
	
	for(i = 0 ; i < ctx->n ; i++) {
			cb->rbcb[i] = ritas_rb_init(ctx, i, 1, cb, RITAS_RB_UPPER_VC, 0);
			if(!cb->rbcb[i])
				goto bad;
	}
	
	return cb;

bad:
	ritas_vc_destroy(ctx, cb);
	return NULL;
}


void ritas_vc_destroy(ritas_t *ctx, ritas_vccb_t *cb) {
	
	int i;

	if(!cb)
		return;

	for(i = 0 ; i < ctx->n ; i++)
		ritas_mbuf_destroy(cb->W[i]);
	free(cb->W);

	ritas_ht_destroy(cb->v_count, (void(*)(void *))ritas_vc_vproc_destroy);

	for(i = 0 ; i < ctx->n ; i++)
		ritas_rb_destroy(ctx, cb->rbcb[i]);
	free(cb->rbcb);
	
	ritas_mvc_destroy(ctx, cb->mvccb);

	if(!cb->upper)
		ritas_aproxy_destroy(ctx);

	free(cb);
}


ritas_vc_vproc_t *ritas_vc_vproc_init(u_short n) {
	
	ritas_vc_vproc_t *vproc;
	
	vproc = malloc(sizeof(*vproc));
	if(!vproc)
		return NULL;

	vproc->proc = malloc(n*sizeof(vproc->proc));
	
	if(!vproc->proc) {
		free(vproc);
		return NULL;
	}
	memset(vproc->proc, '\0', n*sizeof(vproc->proc));
	
	vproc->count = 0;
	
	return vproc;
}

void ritas_vc_vproc_destroy(ritas_vc_vproc_t *vproc) {
	free(vproc->proc);
	free(vproc);
}

int ritas_vc(ritas_t *ctx, u_short cid, u_char *prop, u_short prop_size, u_char *vec, u_short vec_size) {
	
	ritas_vccb_t *cb;
	ritas_mbuf_t *mbuf;
	u_short vsize;
	int n;
	
	pthread_mutex_lock(&ctx->temp_mutex);	
	
	// Init VC control block
	cb = ritas_vc_init(ctx, cid, NULL);
	if(!cb)
		goto bad;
		
	// Init mbuf
	mbuf = ritas_mbuf_init();
	if(!mbuf)
		goto bad;
		
	mbuf->head = mbuf->head - prop_size;
	if(mbuf->head < mbuf->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		goto bad;
	}
	
	memcpy(mbuf->head, prop, prop_size);

	// Send the proposal to the VC layer	
	n = ritas_vc_output(ctx, cb, mbuf);
	if(n == -1)
		goto bad;
	
	mbuf = NULL;
	
	pthread_mutex_unlock(&ctx->temp_mutex);	
	
	// Wait until the VC layer decides
	pthread_mutex_lock(&ctx->aproxy->mutex);
	while(!ctx->aproxy->count)
		pthread_cond_wait(&ctx->aproxy->cond, &ctx->aproxy->mutex);
	pthread_mutex_unlock(&ctx->aproxy->mutex);

	pthread_mutex_lock(&ctx->temp_mutex);

	if(!ctx->aproxy->mvalue) {
		printf("%s\n", ctx->errbuf);
		goto bad;
	}

	vsize = ctx->aproxy->mvalue->tail - ctx->aproxy->mvalue->head;
	if(vsize > vec_size) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): vec_size overflow\n", __func__);
		goto bad;
	}
	
	memcpy(vec, ctx->aproxy->mvalue->head, vsize);
	
	ritas_vc_destroy(ctx, cb);

	pthread_mutex_unlock(&ctx->temp_mutex);
	
	return (int)vsize;

bad:
	if(cb) {
		ritas_mbuf_destroy(mbuf);
		ritas_vc_destroy(ctx, cb);
	}
	return -1;
}

u_short ritas_vc_id(ritas_vccb_t *cb) {
	return cb->id;
}

void *ritas_vc_parent(ritas_vccb_t *cb) {
	return cb->parent;
}
