#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "ritas.h"
#include "ritas_mvc.h"
#include "ritas_ht.h"
#include "ritas_vector.h"

int main() {
	
	ritas_mvc_header_t *mvch;
	u_char *w, *Vbuf, valid;
	u_short *count;
	int i, n;
	ritas_ht_t *vtable = NULL;
	ritas_vector_t *V = NULL;
	u_char *init_vproc_key[4];
	u_char init_vproc_key_size[4];	
	int value;
	int procn, procf;
	u_short value_s;
	
	value = 31337;
	value_s = sizeof(value);

	valid = 0;
	w = NULL;

	procn = 4;
	procf = 1;	
	
	
	for(i = 0 ; i < procn ; i++) {
		if(i != 0) {
			init_vproc_key[i] = malloc(4);
			memcpy(init_vproc_key[i], (u_char *)&value, 4);
			init_vproc_key_size[i] = 4;
			printf("MVC vproc[%d] = %d (%d bytes)\n", i, *(u_int*)init_vproc_key[i], init_vproc_key_size[i]);		
		}
		else {
			init_vproc_key[i] = NULL;
			init_vproc_key_size[i] = 0;
			printf("MVC vproc[%d] = NULL (%d bytes)\n", i, init_vproc_key_size[i]);
		}
	}
	
	if(value_s > 0) {
		vtable = ritas_ht_init(RITAS_MVC_VTABLE_SIZE);
		if(!vtable) {
			printf("%s(): unable to init vtable\n", __func__);
			return -1;
		}
		w = (u_char *)&value;
	}
	else {
		return 1;
	}
	
	V = ritas_vector_init(procn);
	if(!V) {
		printf("%s(): couldn't extract vector\n", __func__);
		goto bad;
	}
	ritas_vector_add(V, (u_char*)&value, sizeof(value));
	ritas_vector_add(V, (u_char*)&value, sizeof(value));
	ritas_vector_add(V, (u_char*)&value, sizeof(value));
	ritas_vector_add(V, NULL, 0);		

	for(i = 0 ; i < V->size ; i++)
		if(V->element_s[i] > 0)
			printf("MVC validating vector[%d] = %d (%d bytes)\n", i, *(u_int*)V->element[i], V->element_s[i]);
		else
			printf("MVC validating vector[%d] = NULL (%d bytes)\n", i, V->element_s[i]);

	
	if(V->size != procn) {
		printf("%s(): vector size != n\n", __func__);
		goto bad;
	}
	
	// TODO: there's a bug in here, revise validation rules
	for(i = 0 ; i < V->size ; i++) {
		if(init_vproc_key_size[i] > 0 && V->element_s[i] > 0 && value_s > 0) {
			if(V->element_s[i] != init_vproc_key_size[i]) {
				ritas_ht_destroy(vtable, NULL);
				return 0;
			}
			if(memcmp(V->element[i], init_vproc_key[i], V->element_s[i])) {
				ritas_ht_destroy(vtable, NULL);
				return 0;
			}
			

			count = ritas_ht_lookup(vtable, V->element[i], V->element_s[i]);
			if(!count) {
				count = malloc(sizeof(*count));
				if(!count) {
					printf("%s(): unable to malloc *count\n", __func__);
					goto bad;
				}
				*count = 0;
				n = ritas_ht_add(vtable, V->element[i], V->element_s[i], count);
				if(n == -1)
					goto bad;
			}
			(*count)++;
		
			if(!valid && *count == procn - 2*procf) {
				if(V->element_s[i] != value_s) {
					ritas_ht_destroy(vtable, NULL);
					return 0;			
				}
				if(memcmp(V->element[i], w, V->element_s[i])) {
					ritas_ht_destroy(vtable, NULL);
					return 0;
				}
				valid = 1;
			}
		}
	}
	
	if(value_s > 0)
		ritas_ht_destroy(vtable, NULL);
	ritas_vector_destroy(V);
	
	printf("VALID: %d\n", valid);
	
	exit(0);
	
bad:
	ritas_ht_destroy(vtable, NULL);
	ritas_vector_destroy(V);
	
	return -1;
}
