#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include "ritas.h"
#include "ritas_rc.h"
#include "ritas_proc.h"
#include "ritas_mbuf.h"
#include "ritas_rb.h"
#include "ritas_vector.h"
#include "ritas_ab.h"
#ifdef MEMWATCH
#include "memwatch.h"
#endif

/*
 * Test suite for RITAS
 */

int main(int argc, char **argv) {
	
	ritas_t *ctx;
	ritas_rbcb_t *rbcb;
	u_char value[256];
	u_short mypid;
	int value_s;
	int n, i;
	u_char errbuf[RITAS_ERRBUF_SIZE];
	u_int proposal;
	u_int decision;
	char pidbuf[64];
	u_char vecbuf[128];
	u_char recv[8];
	ritas_vector_t *vector;
	ritas_abcb_t *abcb;
	ritas_ab_header_t *abh;
	
	if(argc < 2) {
		exit(1);
	}
	
	setlinebuf(stdout);
	mypid = atoi(argv[1]);
	memset(pidbuf, '\0', 64);
	snprintf(pidbuf, 64, "echo %d > pid.%d", getpid(), mypid);
	system(pidbuf);

	ctx = ritas_init(mypid, 4, 1, 4330 + mypid, errbuf); // mypid, n, f
	if(!ctx) {
		fprintf(stderr, "%s\n", errbuf);
		exit(1);
	}
	ritas_proc_add_ipv4(ctx, 0, "127.0.0.1", 4330);
	ritas_proc_add_ipv4(ctx, 1, "127.0.0.1", 4331);
	ritas_proc_add_ipv4(ctx, 2, "127.0.0.1", 4332);
	ritas_proc_add_ipv4(ctx, 3, "127.0.0.1", 4333);

	n = ritas_ab_init(ctx);
	if(n == -1) {
		printf("%s\n", ctx->errbuf);
		ritas_destroy(ctx);
		exit(1);
	}
	
	sleep(3);

	proposal = 31337;
	n = ritas_ab_bcast(ctx, (u_char *)&proposal, sizeof(proposal));
	if(n < 0) {
		ritas_timestamp();
		printf("%s\n", ctx->errbuf);
		exit(1);
	}
	
	for(i = 0 ; i < ctx->n ; i++) {
		n = ritas_ab_recv(ctx, recv, 8);
		ritas_timestamp();
		printf("[%d] received %d bytes\n", ctx->pid, n);
		if(n < 0) {
			printf("%s\n", ctx->errbuf);
			exit(1);
		}
		abh = (ritas_ab_header_t *)recv;
		ritas_timestamp();
		printf("[%d] received msg %d from %d: %d\n", ctx->pid, abh->num, abh->txid, *(u_int *)&recv[4]);
	}
	
	//sleep(3);	

/*

	n = ritas_vc(ctx, 1, (u_char *)&proposal, sizeof(proposal), vecbuf, 128);
	if(n < 0) {
		printf("%s\n", ctx->errbuf);
		exit(1);
	}

	if(n == 0)
		printf("[%d] DECIDED: default\n", mypid);
	else {
		vector = ritas_vector_extract(vecbuf, 128);
		if(!vector) {
			printf("%s: vector error\n", __func__);
			exit(1);
		}
		printf("[%d] DECIDED:\n", mypid);
		for(i = 0 ; i < vector->size ; i++)
			if(vector->element_s[i] > 0)
				printf("vector[%d] = %d (%d bytes)\n", i, *(u_int*)vector->element[i], vector->element_s[i]);
			else
				printf("vector[%d] = NULL (%d bytes)\n", i, vector->element_s[i]);
	}		

	sleep(3);	
*/
/*
	proposal = 31337;
	n = ritas_mvc(ctx, 1, (u_char *)&proposal, sizeof(proposal), &decision, sizeof(decision));
	if(n < 0) {
		fprintf(stderr, "%s\n", ctx->errbuf);
		exit(1);
	}
	
	if(n == 0)
		printf("[%d] DECIDED: default\n", mypid);
	else
		printf("[%d] DECIDED: %d\n", mypid, decision);
	*/

	/* BC
	if(mypid == 0) {
		value[0] = 1;
		value_s = 1;
		ritas_rb_bcast(ctx, 1, value, 1);
	}
	else if(mypid > 0 && mypid < 4) {
		memset(value, '\0', 256);
		if(mypid == 3) sleep(30);
		value_s = ritas_rb_recv(ctx, 0, 1, value, 255);
		if(value_s < 0) {
			exit(1);
		}
		printf("recv'ed: %d\n", value[0]);
	}
	else {
		printf("wrong pid\n");
		exit(1);
	}
	*/
	//sleep(3);
	ritas_timestamp();
	printf("[%d] EXIT\n", mypid);
	ritas_destroy(ctx);
	
}
